package org.n52.swe.sas.bindings;


import java.io.*;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.logging.Logger;

import org.apache.xmlbeans.*;
import org.w3c.dom.Node;

/**
 * @author Jan Torben Heuer <jan.heuer@uni-muenster.de>
 * 
 */
public class XMLBeansParser {
	/* automatically generated Logger */
	@SuppressWarnings("unused")
	private static final Logger LOGGER = Logger.getLogger(XMLBeansParser.class.getName());

	public final static boolean VALIDATE = true;

	public final static boolean DONT_VALIDATE = false;

	public final static XmlOptions PRETTYPRINT = new XmlOptions();
	public final static XmlOptions FAST = new XmlOptions();
	
	static {
		PRETTYPRINT.setSavePrettyPrint();
		PRETTYPRINT.setUseDefaultNamespace();
		FAST.setUseDefaultNamespace();
	}

	public static XmlObject parse(String source) throws InvalidXMLContentException {
		return parse(source, true);
	}

	/**
	 * Reads the given source. The source may be only the xml-document or
	 * contain an application/x-www-form-urlencoded string. In this case, the
	 * request must have the form <em>request=</em>
	 * 
	 * @param source
	 * @return
	 * @throws InvalidXMLContentException
	 */
	public static XmlObject parse(String source, boolean validate) throws InvalidXMLContentException {

		XmlObject doc;
		try {

			doc = XmlObject.Factory.parse(source);
			if (validate) {
				validate(doc);
			}
			return doc;

		} catch (XmlException e) {
			throw new InvalidXMLContentException("Cannot parse xml: "+e.getMessage(), e);
		}
	}

	
	/**
	 * @param resourceAsStream
	 * @throws InvalidXMLContentException
	 */
	public static XmlObject parse(InputStream resourceAsStream) throws InvalidXMLContentException {
		return parse(resourceAsStream, true);
	}

	public static XmlObject parse(InputStream resourceAsStream, boolean validate) throws InvalidXMLContentException {
		XmlObject doc;
		try {
			doc = XmlObject.Factory.parse(resourceAsStream);

			if (validate) {
				validate(doc);
			}
			return doc;

		} catch (XmlException e) {
			
			/* cannot parse xml string. Maybe a stream problem? try to read as String!
			 * This has been implemented because of XmlBeans stream issues. */
			BufferedReader b = new BufferedReader(new InputStreamReader(resourceAsStream));
			
			StringWriter w = new StringWriter();			
			try {
				while(b.ready()) {
					w.write(b.readLine());
				}
			} catch (IOException e2) {
				throw new InvalidXMLContentException("Cannot read the document: Transmission interrupted!", e);
			}
			
			try {
				return XmlObject.Factory.parse(w.toString());
			} catch (XmlException e1) {
				throw new InvalidXMLContentException("The document you supplied was incomplete. Please try again.", e);
			}
			
		} catch (IOException e) {
			throw new InvalidXMLContentException("Cannot read the document: Transmission interrupted!", e);
		}
	}

	public static XmlObject parse(Node xmlnode, boolean validate) throws InvalidXMLContentException {
		XmlObject doc;
		try {
			doc = XmlObject.Factory.parse(xmlnode);

			if (validate) {
				validate(doc);
			}
			return doc;

		} catch (XmlException e) {
			throw new InvalidXMLContentException("Cannot read the node", e);
		}
	}
	public static String ObjectToString(XmlObject doc) throws InvalidXMLContentException {
		return ObjectToString(doc, true);
	}

	public static String ObjectToString(XmlObject doc, boolean validate) throws InvalidXMLContentException {
		return ObjectToString(doc, validate, false);
	}

	public static String ObjectToString(XmlObject doc, boolean validate, boolean prettyprint) throws InvalidXMLContentException {
		if (validate) {
			validate(doc);
		}
		if (prettyprint) {
			return doc.xmlText(PRETTYPRINT);
		} else {
			return doc.xmlText(FAST);
		}
	}

	public static void writeObject(XmlObject doc, PrintWriter out) throws IOException {
		doc.save(out, PRETTYPRINT);
	}

	public static void writeObject(XmlObject doc, PrintWriter out, boolean validate) throws IOException, InvalidXMLContentException {
		if (validate) {
			validate(doc);
		}
		writeObject(doc, out);
	}

	/**
	 * Validates an Xmldoc. If the validation fails, the exception contains a
	 * detailed list of errors.
	 * 
	 * @param doc
	 * @throws InvalidXMLContentException
	 */
	public static void validate(XmlObject doc) throws InvalidXMLContentException {
		ArrayList validationErrors = new ArrayList();
		XmlOptions validationOptions = new XmlOptions();
		validationOptions.setErrorListener(validationErrors);

		boolean isValid = doc.validate(validationOptions);

		// create XmlException with error-message if the XML is invalid.
		if (!isValid) {

			throw new InvalidXMLContentException("Invalid xml content\n", validationErrors);
		}
	}

	/**
	 * Extracts an element from an xmldoc. This is only useful for abstract or
	 * choice types. It is done via reflection
	 * 
	 * @param doc
	 *            The xml elemet
	 * @param element
	 *            the name of the Method, for example <code>getUom</code>
	 * @return the returnvalue of the Method or null if the Method does not exist.
	 * @throws InvalidXMLContentException 
	 */
	public static String getElement(XmlObject doc, String element) throws InvalidXMLContentException {
		try {
			for (Method method : doc.getClass().getMethods()) {
				if (method.getName().equals(element)) {
					return method.invoke(doc, (Object[]) null).toString();
				}
			}
			
			return null;
		} catch (Exception e) {
			throw new InvalidXMLContentException("cannot extract "+  element +" from " + doc.schemaType().getName().toString(),e);
		}

	}

}
