/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml;

import java.util.Date;
import java.util.logging.Logger;

import net.opengis.eml.x001.EMLDocument.EML;
import net.opengis.sensorML.x101.*;
import net.opengis.swe.x101.DataComponentPropertyType;
import de.ifgi.lehre.thesisEverding.eml.esper.EsperController;
import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.process.InputDescription;
import de.ifgi.lehre.thesisEverding.process.OutputDescription;
import de.ifgi.lehre.thesisEverding.sml.SensorMLParsingHelper;


/**
 * Representation of an EML process.
 * 
 * @author Thomas Everding
 * 
 */
public class EMLProcess extends AEMLProcess {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	/**
	 * Controller to execute EML via esper.
	 */
	private EsperController esperController;
	
	/**
	 * EML specifying the event patterns
	 */
	private EML eml;
	
	
	@Override
	public void initialize(AbstractProcessType aProcess) {
		//check process type
		if (!(aProcess instanceof ProcessModelType)) {
			//wrong process type
			logger.severe("Process type is not supported.");
			return;
		}
		
		ProcessModelType pm = (ProcessModelType) aProcess;
		
		//load EML TODO
//		try {
//			if (pm.getMethod().isSetProcessMethod()
//				&& pm.getMethod().getProcessMethod().isSetAlgorithm()
//				&& pm.getMethod().getProcessMethod().getAlgorithm().getAlgorithmDefinition().isSetEml()) {
//				//get EML from SensorML
//				String emlString = pm.getMethod().getProcessMethod().getAlgorithm().getAlgorithmDefinition().getEml().toString(); 
//				EMLDocument emlDoc;
//				emlDoc = EMLDocument.Factory.parse(emlString);
//				eml = emlDoc.getEML();
//			}
//			else {
//				logger.fatal("SensorML document is not valid!");
//				return;
//			}
//		}
//		catch (Throwable t) {
//			logger.error(t.getMessage(), t);
//			return;
//		}
		
		//build input descriptions
		InputDescription[] inputDescriptions;
		if (pm.isSetInputs()
			&& pm.getInputs().isSetInputList()) {
			//build description for every input
			IoComponentPropertyType[] components = pm.getInputs().getInputList().getInputArray();
			inputDescriptions = new InputDescription[components.length];
			InputDescription descr;
			
			for (int i = 0; i < components.length; i++) {
				descr = new InputDescription();
				SensorMLParsingHelper.loadDescription(components[i], descr);
				inputDescriptions[i] = descr;
			}
		}
		else {
			inputDescriptions = new InputDescription[0];
		}
		
		//build parameter descriptions
		InputDescription[] parameterDescriptions;
		if (pm.isSetParameters()
			&& pm.getParameters().isSetParameterList()) {
			//build description for every parameter (parameters with value are not supported)
			DataComponentPropertyType[] components = pm.getParameters().getParameterList().getParameterArray();
			parameterDescriptions = new InputDescription[components.length];
			InputDescription descr;
			
			for (int i = 0; i < components.length; i++) {
				descr = new InputDescription();
				SensorMLParsingHelper.loadDescription(components[i], descr);
				parameterDescriptions[i] = descr;
			}
		}
		else {
			parameterDescriptions = new InputDescription[0];
		}
		
		//build output descriptions
		OutputDescription[] outputDescriptions;
		if (pm.isSetOutputs()
			&& pm.getOutputs().isSetOutputList()) {
			//build description for every output
			IoComponentPropertyType[] components = pm.getOutputs().getOutputList().getOutputArray();
			outputDescriptions = new OutputDescription[components.length];
			OutputDescription descr;
			
			for (int i = 0; i < components.length; i++) {
				descr = new OutputDescription();
				SensorMLParsingHelper.loadDescription(components[i], descr);
				outputDescriptions[i] = descr;
			}
		}
		else {
			outputDescriptions = new OutputDescription[0];
		}
		
		//build esper controller
		this.esperController = new EsperController(this);
		
		//start listening,...
		this.esperController.initialize(eml, inputDescriptions, parameterDescriptions, outputDescriptions);
	}
	

	@Override
	public void handleInput(String inputName, Object value) {
		//build event
		MapEvent event;
		
		if (value instanceof MapEvent) {
			//value can be pushed directly
			event = (MapEvent) value;
		}
		else {
			//build new MapEvent
			Date now = new Date();
			event = new MapEvent(now.getTime(), now.getTime());
			
			//set value
			event.put(MapEvent.VALUE_KEY, value);
		}
		
		//send event
//		logger.debug("");
//		logger.debug("pushing event");
		this.esperController.sendEvent(inputName, event);
	}


	@Override
	public void stop() {
		this.esperController.stop();
	}
	
	
	/**
	 * performs the output of a value
	 * 
	 * @param outputName name of the output
	 * @param value value of the output
	 */
	public void doOutput(String outputName, Object value) {
		this.parent.handleOutput(this.referencePrefix + outputName, value);
	}
}
