/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml;

import java.util.HashMap;
import java.util.Vector;
import java.util.logging.Logger;

import net.opengis.sensorML.x101.*;
import net.opengis.sensorML.x101.ComponentsDocument.Components.ComponentList.Component;
import net.opengis.sensorML.x101.ConnectionDocument.Connection;
import net.opengis.sensorML.x101.LinkDocument.Link;

import com.espertech.esper.collection.Pair;


/**
 * Representation of an EML process chain.
 * 
 * @author Thomas Everding
 *
 */
public class EMLProcessChain extends AEMLProcess implements IEMLProcessParent{
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private static final String PREFIX_INPUT = "this/inputs/";
	
	private static final String PREFIX_PARAMETER = "this/parameter/";
	
	private static final String PREFIX_OUTPUT = "this/outputs/";
	
	/**
	 * contains all sub processes (components) by name
	 */
	private HashMap<String, AEMLProcess> components;
	
	/**
	 * contains all connections
	 * 
	 * key (string): source name
	 * value: Vector<pair>
	 * 		pair.first: {@link AEMLProcess} as receiver (handleInput)
	 * 		pair.second: {@link String} name of the input
	 */
	private HashMap<String, Vector<Pair<AEMLProcess, String>>> connections;
	
	
	/**
	 * 
	 * Constructor
	 *
	 */
	public EMLProcessChain() {
		this.components = new HashMap<String, AEMLProcess>();
		this.components.put("this", this);
		
		this.connections = new HashMap<String, Vector<Pair<AEMLProcess,String>>>();
	}

	@Override
	public void initialize(AbstractProcessType aProcess) {
		//check process type
		if(!(aProcess instanceof ProcessChainType)) {
			logger.severe("Process type is not supported");
			return;
		}
		
		ProcessChainType pc = (ProcessChainType) aProcess;
		
		//build and initialize every sub abstract process (model or chain)
		Component[] components = pc.getComponents().getComponentList().getComponentArray();
		for (Component comp : components) {
			//add as sub process
			this.addSubProcess(comp);
		}
		
		//parse connections
		Connection[] connections = pc.getConnections().getConnectionList().getConnectionArray();
		for (Connection conn : connections) {
			//add connection
			this.addConnection(conn);
		}
	}


	public void handleOutput(String outputName, Object value) {
		//check for connection
		if (!this.connections.containsKey(outputName)) {
			logger.severe("no connection found to handle output from '" + outputName + "'.");
			return;
		}
		
		Vector<Pair<AEMLProcess, String>> processesAndNames = this.connections.get(outputName);
//		Pair<AEMLProcess, String> processAndName = this.connections.get(outputName);
		
		//loop
		for (Pair<AEMLProcess, String> processAndName : processesAndNames) {
			//check for internal or external communication
			if (processAndName.getFirst() == this) {
				//external: send to parent of this
				this.parent.handleOutput(PREFIX_OUTPUT + processAndName.getSecond(), value);
			}
			else {
				//internal: send to another sub process of this
				processAndName.getFirst().handleInput(processAndName.getSecond(), value);
			}
		}
	}

	@Override
	public void handleInput(String inputName, Object value) {
		/*
		 * get connection for input
		 */
		Vector<Pair<AEMLProcess, String>> processesAndNames;
//		Pair<AEMLProcess, String> processAndName;
		boolean isInput;
		//try as input
		if (this.connections.containsKey(PREFIX_INPUT + inputName)) {
			//found as input
			processesAndNames = this.connections.get(PREFIX_INPUT + inputName);
			isInput = true;
		}
		
		//try as parameter
		else if (this.connections.containsKey(PREFIX_PARAMETER + inputName)) {
			//found as parameter
			processesAndNames = this.connections.get(PREFIX_PARAMETER + inputName);
			isInput = false;
		}
		
		else {
			//not found
			logger.severe("No connection for '" + inputName + "' found.");
			return;
		}
		
		/*
		 * handle input
		 */
		
		//loop
		for (Pair<AEMLProcess, String> processAndName : processesAndNames) {
			if (processAndName.getFirst() == this) {
				//input is directly connected to output
				if (isInput) {
					this.handleOutput(PREFIX_INPUT + inputName, value);
				}
				else {
					this.handleOutput(PREFIX_PARAMETER + inputName, value);
				}
			}
			else {
				//send input to sub component
				processAndName.getFirst().handleInput(processAndName.getSecond(), value);
			}
		}
	}
	
	
	/**
	 * Adds a process
	 * 
	 * @param comp SensorML description of the process
	 */
	private void addSubProcess(Component comp) {
		//create sub component
		String componentName = comp.getName();
		AEMLProcess process = null;
		if (comp.getProcess() instanceof ProcessChainType) {
			//create sub chain
			process = new EMLProcessChain();
		}
		else if (comp.getProcess() instanceof ProcessModelType) {
			//create sub process
			process = new EMLProcess();
		}
		
		//initialize component
		process.setParent(this);
		process.setReferencePrefix(componentName + "/outputs/");
		process.initialize(comp.getProcess());
		
		//store
		this.components.put(componentName, process);
	}
	
	
	/**
	 * Adds a connection between two (sub)processes
	 * 
	 * @param conn description of the connection
	 */
	private void addConnection(Connection conn) {
		//only Links allowed
		if (!conn.isSetLink()) {
			logger.severe("Connection must be specified as link");
			return;
		}
		
		//get references
		Link link = conn.getLink();
		String source = link.getSource().getRef(); //used as key in the connections map
		String destination = link.getDestination().getRef();
		
		/*
		 * parse destination reference
		 * 
		 * "process/in, out or param/name"
		 */
		String destinationName;
		String processName;
		AEMLProcess destinationProcess;
		String[] segments = destination.split("/");
		
		if (!(segments.length == 3)) {
			logger.severe("Connection reference has wrong syntax. (" + destination + ")");
			return;
		}
		
		//get process
		processName = segments[0];
		if (!this.components.containsKey(processName)) {
			logger.severe("Process (" + processName + ") not found");
			return;
		}
		destinationProcess = this.components.get(processName);
		
		//get name (of input / parameter)
		destinationName = segments[2];
		
		//store connection
		if (!this.connections.containsKey(source)) {
			//create new connection vector for source name
			this.connections.put(source, new Vector<Pair<AEMLProcess,String>>());
		}
		this.connections.get(source).add(new Pair<AEMLProcess, String>(destinationProcess, destinationName));
//		this.connections.put(source, new Pair<AEMLProcess, String>(destinationProcess, destinationName));
	}

	@Override
	public void stop() {
		//stop all components
		for (AEMLProcess comp : this.components.values()) {
			comp.stop();
		}
	}
}
