/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml;

import java.util.*;
import java.util.logging.Logger;

import net.opengis.sensorML.x101.*;

import com.espertech.esper.collection.Pair;

import de.ifgi.lehre.thesisEverding.eml.constants.SupportedDataTypes;
import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.process.*;
import de.ifgi.lehre.thesisEverding.processCommunication.*;


/**
 * Executes EML on esper
 * 
 * @author Thomas Everding
 * 
 */
public class EMLProcessExecutor extends AProcessExecutor implements IEMLProcessParent, IInputReceiver {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private EMLProcessPlugin plugin;
	
	private AbstractProcessType aProcess;
	
	private Vector<NonBufferedListener> listeners;
	
	private Pair<String, Object> newestValue;
	
	/**
	 * Process to execute. Can be a simple process or a process chain.
	 */
	private AEMLProcess emlProcess;
	

	/**
	 * 
	 * Constructor
	 * 
	 * @param plugin Process Plugin
	 * @param aProcess SensorML AbstractProcessType
	 */
	public EMLProcessExecutor(EMLProcessPlugin plugin, AbstractProcessType aProcess) {
		this.plugin = plugin;
		this.aProcess = aProcess;
		
		this.listeners = new Vector<NonBufferedListener>();
	}
	

	@Override
	protected void doCalculation() {
		//build new event
		MapEvent event;
		
		if (this.newestValue.getSecond() instanceof MapEvent) {
			//event received
			event = (MapEvent) this.newestValue.getSecond();
		}
		else {
			//value received
			Date now = new Date();
			event = new MapEvent(now.getTime(), now.getTime());
			
			//set value
			event.put(MapEvent.VALUE_KEY, this.newestValue.getSecond());
		}
		
		//send event to esper engine
		this.emlProcess.handleInput(this.newestValue.getFirst(), event);
	}
	

	@Override
	protected void startExecution() {
		//check process type
		if (this.aProcess instanceof ProcessModelType) {
			//only a simple process model
			this.emlProcess = new EMLProcess();
		}
		else if (this.aProcess instanceof ProcessChainType) {
			//process chain
			this.emlProcess = new EMLProcessChain();
		}
		else {
			//not supported
			logger.severe("process type is not supported");
			return;
		}
		
		//initialize process
		this.emlProcess.setParent(this);
		this.emlProcess.setReferencePrefix("");
		this.emlProcess.initialize(this.aProcess);
		
		//start input listeners
		this.startListening();
	}
	

	@Override
	protected void stopExecution() {
		running = false;
		
		//stop input listeners
		for (NonBufferedListener lis : this.listeners) {
			lis.setRunning(false);
		}
		this.listeners = new Vector<NonBufferedListener>();
		
		//stop esper
		this.emlProcess = null;
	}
	

	@SuppressWarnings("unchecked")
	public void handleOutput(String outputName, Object value) {
		//Check name
		if (outputName.contains("/")) {
			//cut of reference part
			outputName = outputName.substring(outputName.lastIndexOf("/") + 1);
		}
		
		/*
		 * perform output
		 */
		
		//get connectors and description
		Vector<IOutputConnector> connectors = this.outputConnectors.get(outputName);
		
		OutputDescription[] descriptions = this.plugin.getOutputDescriptions();
		OutputDescription descr = null;
		for (int i = 0; i < descriptions.length; i++) {
			descr = descriptions[i];
			
			if (descr.getName().equals(outputName)) {
				//description found
				break;
			}
		}
		
		//check description
		if (descr == null) {
			//no output description found
			logger.severe("no description for '" + outputName + "' found");
			return;
		}
		
		//check output data type
		boolean sendMapEvent;
		if (descr.getDataType().equals(SupportedDataTypes.EVENT)) {
			sendMapEvent = true;
		}
		else {
			sendMapEvent = false;
		}
		
		//build output value
		Object outputValue;
		if (value instanceof Map) {
			//output is of type map (MapEvent implements Map)
			if (sendMapEvent) {
				//parameter "value" is of correct type
				outputValue = value;
			}
			else {
				//get the value for the output from parameter "value"
				Map map = (Map) value;
				
				outputValue = map.get(MapEvent.VALUE_KEY);
				//other keys not allowed here, select a different property via EML if needed
			}
		}
		else {
			//output is of another type
			if (sendMapEvent) {
				//build map event with parameter "value" as value
				Date now = new Date();
				MapEvent event = new MapEvent(now.getTime(), now.getTime());
				event.put(MapEvent.VALUE_KEY, value);
				
				outputValue = event;
			}
			else {
				//output may be of correct type
				outputValue = value;
			}
		}
		
		//perform output
		if (connectors != null) {
			for (IOutputConnector conn : connectors) {
				conn.write(outputValue);
			}
		}
	}
	

	public synchronized void receiveInput(String inputName, Object value) {
		this.newestValue = new Pair<String, Object>(inputName, value);
		
		//calculate the received value
		this.doCalculation();
	}
	

	@Override
	public synchronized void calculate() {
		//no need for this method in this plugin / executor
		return;
	}
	

	/**
	 * builds and starts the input listeners
	 */
	private void startListening() {
		/*
		 * build listeners
		 */
		NonBufferedListener listener;
		IInputConnector conn;
		
		//listeners for inputs
		InputDescription[] descriptions = this.plugin.getInputDescriptions();
		
		for (InputDescription descr : descriptions) {
			conn = this.inputConnectors.get(descr.getName());
			listener = new NonBufferedListener(descr.getName(), conn, this);
			this.listeners.add(listener);
		}
		
		//listeners for parameters
		descriptions = this.plugin.getParameterDescriptions();
		
		for (InputDescription descr : descriptions) {
			conn = this.inputConnectors.get(descr.getName());
			listener = new NonBufferedListener(descr.getName(), conn, this);
			this.listeners.add(listener);
		}
		
		/*
		 * start listeners
		 */
		Thread t;
		for (NonBufferedListener lis : this.listeners) {
			t = new Thread(lis);
			t.setDaemon(true);
			t.start();
		}
	}
}