/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml.esper;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Logger;

import net.opengis.eml.x001.EMLDocument.EML;

import com.espertech.esper.client.Configuration;
import com.espertech.esper.client.EPServiceProvider;
import com.espertech.esper.client.EPServiceProviderManager;
import com.espertech.esper.client.EPStatement;

import de.ifgi.lehre.thesisEverding.eml.constants.SupportedDataTypes;
import de.ifgi.lehre.thesisEverding.eml.EMLProcess;
import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.eml.pattern.APattern;
import de.ifgi.lehre.thesisEverding.eml.pattern.PatternComplex;
import de.ifgi.lehre.thesisEverding.eml.pattern.PatternRepetitive;
import de.ifgi.lehre.thesisEverding.eml.pattern.SelFunction;
import de.ifgi.lehre.thesisEverding.eml.pattern.Statement;
import de.ifgi.lehre.thesisEverding.process.InputDescription;
import de.ifgi.lehre.thesisEverding.process.OutputDescription;


/**
 * central class for executing a set of esper EPL statements for a single process
 * 
 * @author Thomas Everding
 * 
 */
public class EsperControllerBackup {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private Configuration config;
	
	private EPServiceProvider epService;
	
	private HashMap<String, StatementListener> listeners;
	
	private HashMap<String, CountingListener> countingListeners;
	
	private HashMap<String, EPStatement> epStatements;
	
	private EMLParser parser;
	
	private OutputDescription[] outputDescriptions;
	
	private HashMap<String, Object> inputEventDataTypes;
	
	private EMLProcess process;

	private InputDescription[] inputDescriptions;
	
	
	/**
	 * 
	 * Constructor
	 * 
	 * @param process the (SensorML)process this controller belongs to
	 * 
	 */
	public EsperControllerBackup(EMLProcess process) {
		this.process = process;
		
		this.config = new Configuration();
		this.listeners = new HashMap<String, StatementListener>();
		this.countingListeners = new HashMap<String, CountingListener>();
		//		this.timerListeners = new HashMap<String, TimerListener>();
		this.epStatements = new HashMap<String, EPStatement>();
		this.inputEventDataTypes = new HashMap<String, Object>();
	}
	

	/**
	 * Registers an event with a name at the engine. Nestable maps are allowed.
	 * 
	 * @param eventName Name of the event used in the patterns (mostly the inputName).
	 * @param eventProperties Properties of the event. Each event is send as an HashMap. These Properties must
	 * contain an entry for each key in the HashMap containing the data type of the HashMap value.
	 */
	public synchronized void registerEvent(String eventName, Map<String, Object> eventProperties) {
		this.config.addNestableEventTypeAlias(eventName, eventProperties);
	}
	

	/**
	 * Registers an event for the given {@link InputDescription}
	 * 
	 * @param descr description of an process input
	 */
	private void registerEvent(InputDescription descr) {
		String eventName = descr.getName();
		HashMap<String, Object> props = new HashMap<String, Object>();
		
		if (descr.getDataType() != SupportedDataTypes.EVENT) {
			/*
			 * check data type
			 */

			if (descr.getDataType().equals(SupportedDataTypes.BOOLEAN)) {
				//boolean
				props.put(MapEvent.VALUE_KEY, Boolean.class);
			}
			
			else if (descr.getDataType().equals(SupportedDataTypes.CATEGORY)) {
				//category
				props.put(MapEvent.VALUE_KEY, String.class);
			}
			
			else if (descr.getDataType().equals(SupportedDataTypes.COUNT)) {
				//count
				props.put(MapEvent.VALUE_KEY, Integer.class);
			}
			
			else if (descr.getDataType().equals(SupportedDataTypes.QUANTITY)) {
				//quantity
				props.put(MapEvent.VALUE_KEY, Double.class);
			}
			
			else if (descr.getDataType().equals(SupportedDataTypes.TEXT)) {
				//text
				props.put(MapEvent.VALUE_KEY, String.class);
			}
			
			else if (descr.getDataType().equals(SupportedDataTypes.TIME)) {
				//time
				props.put(MapEvent.VALUE_KEY, Long.class);
			}
			else {
				//unknown
				props.put(MapEvent.VALUE_KEY, Object.class);
			}
			
			//save input event data type
			this.inputEventDataTypes.put(eventName, props.get(MapEvent.VALUE_KEY));
		}
		else {
			/*
			 * register swe:Event
			 */
			this.buildEventPropertyDataTypeMap(props, descr);
			
			//save input event data type
			this.inputEventDataTypes.put(eventName, props);
		}
		
		//register at esper engine
		this.registerEvent(eventName, props);
	}
	

	/**
	 * builds the properties map for events (swe:Event)
	 * 
	 * @param props data type map of the input
	 * @param description description of the connection delivering an swe:Event
	 */
	private void buildEventPropertyDataTypeMap(HashMap<String, Object> props, InputDescription description) {
		//parse swe:Event data types from description
		HashMap<String, Object> dataTypes = EventDataTypeParser.parse(description);
		
		//add data types to properties
		for (String key : dataTypes.keySet()) {
			props.put(key, dataTypes.get(key));
		}
	}


	/**
	 * Initializes the controller with the EML patterns.
	 * 
	 * @param eml The event patterns in EML.
	 * @param inputDescriptions descriptions of the inputs of the process
	 * @param paramDescriptions descriptions of the not set parameters of the process
	 * @param outputDescriptions descriptions of the outputs of the process
	 */
	public void initialize(EML eml, InputDescription[] inputDescriptions, InputDescription[] paramDescriptions,
			OutputDescription[] outputDescriptions) {
//		parser = new EMLParser(this);
		parser.parseEML(eml);
		HashMap<String, APattern> patterns = parser.getPatterns();
		
		this.outputDescriptions = outputDescriptions;
		this.inputDescriptions = inputDescriptions;
		
		//register events for process inputs
		for (InputDescription descr : inputDescriptions) {
			this.registerEvent(descr);
		}
		
		//register events for not set process parameters
		for (InputDescription descr : paramDescriptions) {
			this.registerEvent(descr);
		}
		
		//register custom guards
		this.registerCustomGuards();
		
		//register custom functions
		this.registerCustomFunctions();
		
		//build listeners
		this.buildListeners(patterns);
		
		//TODO entfernen
		this.logStatements();
		
		//initialize esper
		this.epService = EPServiceProviderManager.getDefaultProvider(this.config);
		
		//initialize repetitive count listeners
		this.initializeCountingListeners();
		
		//initialize listeners
		this.initializeListeners();
		
		//register debug listeners
//		this.buildDebugListeners();
		
		//		//initialize timer listeners
		//		this.initializeTimerListeners();
		//		
		//		//publish one internal timer event for every timer pattern (start these patterns)
		//		this.startTimerPatterns();
		
		//send initial counting event
		this.sendInitialCountingEvent();
	}
	

	/**
	 * registers the namespace of the custom functions
	 */
	private void registerCustomFunctions() {
		//standard namespace
		this.config.addImport("de.ifgi.lehre.thesisEverding.eml.esper.customFunctions.*");
	}


	/**
	 * logs all created statements
	 */
	private void logStatements() {
		logger.fine("Statements:");
		for (String key : this.listeners.keySet()) {
			logger.fine("\t" + key);
		}
		logger.fine("--");
	}


	/**
	 * registers the custom guards
	 */
	private void registerCustomGuards() {
//		//greater than
//		this.config.addPlugInPatternGuard(Constants.GUARD_COMPARISON_NAMESPACE, 
//										  "greater", 
//										  "de.ifgi.lehre.thesisEverding.eml.complexGuard.comparison.GreaterThanGuardFactory");
	}


	/**
	 * registers debug listeners for every statement
	 */
	@SuppressWarnings("unused")
	private void buildDebugListeners() {
		//a debug listener for every statement
		EPStatement statement;
		
		for (String key : this.epStatements.keySet()) {
			statement = this.epStatements.get(key);
			
			statement.addListener(new DebugListener());
		}
	}
	

	/**
	 * sends an initial counting event to all counting listeners
	 */
	private void sendInitialCountingEvent() {
		CountingListener cListener;
		for (String key : this.countingListeners.keySet()) {
			cListener = this.countingListeners.get(key);
			
			//publish start event for counting
			Date now = new Date();
			MapEvent event = new MapEvent(now.getTime(), now.getTime());
			event.put(MapEvent.VALUE_KEY, 1);
			
			this.sendEvent(cListener.getInputEventName(), event);
		}
	}
	

	/**
	 * Sends an event to the esper runtime.
	 * 
	 * @param name The name of the event.
	 * @param event The event itself.
	 */
	public synchronized void sendEvent(String name, MapEvent event) {
//				logger.debug("");
//				logger.debug("posting new event:");
//				logger.debug("\tname:  " + name);
//				logger.debug("\tevent: " + event);
//				for (String key : event.keySet()) {
//					logger.debug("\t" + key + ": \t" + event.get(key));
//				}
//				logger.debug("");
		this.epService.getEPRuntime().sendEvent(event, name);
	}
	

	/**
	 * builds and registers the listeners for all patterns
	 * 
	 * @param patterns the patterns
	 */
	private void buildListeners(HashMap<String, APattern> patterns) {
		HashMap<String, Object> completedPatterns = new HashMap<String, Object>();
		Vector<APattern> uncompletedPatterns = new Vector<APattern>();
		APattern patt;
		
		//check every pattern
		for (String key : patterns.keySet()) {
			patt = patterns.get(key);
			
			//first run: only simple and timer patterns
			if (patt instanceof PatternComplex || patt instanceof PatternRepetitive) {
				//these patterns need other patterns to be registered first
				uncompletedPatterns.add(patt);
				continue;
			}
			
			this.buildListenersForPattern(patt);
			completedPatterns.put(patt.getPatternID(), patt);
		}
		
		//second run: complex and repetitive patterns
		int i = -1;
		int maxTests = uncompletedPatterns.size() * 3;
		while (uncompletedPatterns.size() > 0) {
			uncompletedPatterns = this.buildComplexListeners(completedPatterns, uncompletedPatterns);
			
			//check for loop
			i++;
			if (i > maxTests) {
				logger.severe("One of the patterns can not be build or there is a loop in the patterns. This is not allowed.");
				break;
			}
		}
	}
	

	/**
	 * builds the listeners for patterns that use other patterns (complex and repetitive)
	 * 
	 * @param completedPatterns already completed patterns
	 * @param uncompletedPatterns patterns without listener
	 * @return the patterns witch could not be build
	 */
	private Vector<APattern> buildComplexListeners(HashMap<String, Object> completedPatterns,
			Vector<APattern> uncompletedPatterns) {
		
		Vector<APattern> stillUncomPatterns = new Vector<APattern>();
		PatternComplex cp;
		PatternRepetitive rp;
		for (APattern pat : uncompletedPatterns) {
			//check if all internal patterns are already completed
			if (pat instanceof PatternComplex) {
				//complex pattern
				cp = (PatternComplex) pat;
				
				if (completedPatterns.containsKey(cp.getFirstPatternID())
						&& completedPatterns.containsKey(cp.getSecondPatternID())) {
					//build pattern listeners
					this.buildListenersForPattern(pat);
					completedPatterns.put(pat.getPatternID(), pat);
				}
				else {
					//append to list, try later
					stillUncomPatterns.add(pat);
				}
			}
			else {
				//repetitive pattern
				rp = (PatternRepetitive) pat;
				
				if (completedPatterns.containsKey(rp.getPatternToRepeatID())) {
					//build pattern listeners
					this.buildListenersForPattern(pat);
					completedPatterns.put(pat.getPatternID(), pat);
				}
				else {
					//append to list, try later
					stillUncomPatterns.add(pat);
				}
			}
		}
		
		return stillUncomPatterns;
	}
	

	/**
	 * builds and registers the listeners for a single pattern
	 * 
	 * @param pattern the pattern
	 */
	private void buildListenersForPattern(APattern pattern) {
		if (pattern instanceof PatternRepetitive) {
			/*
			 * repetitive pattern needs two statements per select function
			 * 
			 * first statement is the counting statement, the others are the selecting statements.
			 */
			Statement[] statements = pattern.createEsperStatements();
			
			//create new CountingListener for the first statement
//			CountingListener cListener = new CountingListener(this,
//					((PatternRepetitive) pattern).getInputEventName());
//			this.countingListeners.put(statements[0].getStatement(), cListener);
			
			//create listeners for the selecting statements
			for (int i = 1; i < statements.length; i++) {
				this.buildListener(statements[i]);
			}
			
		}
		else {
			/*
			 * other pattern only needs one per statement
			 */
			for (Statement statement : pattern.createEsperStatements()) {
				this.buildListener(statement);
			}
		}
	}
	

	/**
	 * builds the listener for a single statement
	 * 
	 * @param statement the statement
	 */
	private void buildListener(Statement statement) {
		@SuppressWarnings("unused")
		StatementListener listener;
		//create new listener
//		listener = new StatementListener(statement, this);
		
		//add listener to map
//		listeners.put(statement.getStatement(), listener);
	}
	

	/**
	 * initializes the listeners (registers them at the statements)
	 */
	private synchronized void initializeListeners() {
		//for every statement in the map
		EPStatement epStatement;
		for (String statement : listeners.keySet()) {
			//register statements at engine
			epStatement = this.epService.getEPAdministrator().createEPL(statement);
			
			//register listener at esper statement
			epStatement.addListener(listeners.get(statement));
			
			//store epStatements
			this.epStatements.put(statement, epStatement);
		}
	}
	

	/**
	 * initializes the counting listeners
	 */
	private synchronized void initializeCountingListeners() {
		//for every statement in the map
		EPStatement epStatement;
		CountingListener cListener;
		for (String statement : this.countingListeners.keySet()) {
			cListener = this.countingListeners.get(statement);
			
			//register statement at engine
			epStatement = this.epService.getEPAdministrator().createEPL(statement);
			
			//register listener at esper statement
			epStatement.addListener(cListener);
			
			//store epStatements
			this.epStatements.put(statement, epStatement);
		}
	}
	

	/**
	 * Searches for the data type of a property.
	 * 
	 * @param fullPropertyName the full EML name of the property
	 * 
	 * @return a java.lang.Class or a Map containing Classes and/or further Maps
	 */
	public Object getDatatype(String fullPropertyName) {
		//split into event and property name part
		String eventName;
		String propertyName;
		int lastSlash = fullPropertyName.lastIndexOf("/");
		
		propertyName = fullPropertyName.substring(lastSlash + 1);
		
		int lastButOneSlash = fullPropertyName.substring(0, lastSlash).lastIndexOf("/");
		
		if (lastButOneSlash <= 0) {
			//full name looks like "event/value"
			eventName = fullPropertyName.substring(0, lastSlash);
		}
		else {
			//full name looks like "event/nestedEvent/value", we need nestedEvent
			eventName = fullPropertyName.substring(lastButOneSlash + 1, lastSlash);
		}
		
		//check all inputs first
		for (InputDescription descr : this.inputDescriptions) {
			if (descr.getName().equals(eventName)) {
				return DataTypeNameToClassConverter.convert(descr.getDataType());
			}
		}
		
		//then check all patterns
		for (APattern pat : parser.getPatterns().values()) {
			for (SelFunction sel : pat.getSelectFunctions()) {
				if (sel.getNewEventName().equals(eventName)) {
					//this select function defines the data type
					return sel.getDataTypes().get(propertyName);
				}
			}
		}
		return null;
	}
	

	/**
	 * get a map containing all data types of an event
	 * 
	 * @param eventName name of the event (only the event name)
	 * 
	 * @return a map containing all data types of an event 
	 * or the class of the data type if the event is an input event
	 */
	public Object getEventDatatype(String eventName) {
		//check input data types
		if (this.inputEventDataTypes.containsKey(eventName)) {
			//event is process input
			return this.inputEventDataTypes.get(eventName);
		}
		
		//check pattern outputs
		for (APattern pat : this.parser.getPatterns().values()) {
			for (SelFunction sel : pat.getSelectFunctions()) {
				if (sel.getNewEventName().equals(eventName)) {
					//event definition found
					return sel.getDataTypes();
				}
			}
		}
		
		//nothing found
		logger.warning("No data type description found for event '" + eventName + "'.");
		return null;
	}
	

	/**
	 * 
	 * @param name name of the output
	 * @return the {@link OutputDescription} oft an output or <code>null</code> if there is no output for the
	 * name
	 */
	public OutputDescription getOutputDescription(String name) {
		for (OutputDescription descr : this.outputDescriptions) {
			if (descr.getName().equals(name)) {
				//output description found
				return descr;
			}
		}
		logger.severe("No description for output '" + name + "' found.");
		return null;
	}
	

	/**
	 * stops the esper engine
	 */
	public void stop() {
		this.epService.destroy();
	}
	

	/**
	 * Performs the output of a value.
	 * 
	 * @param outputName the name of the output
	 * @param value the value to send
	 */
	public void doOutput(String outputName, Object value) {
		this.process.doOutput(outputName, value);
	}
	

	//	/**
	//	 * @return the epService
	//	 */
	//	public EPServiceProvider getEpService() {
	//		return this.epService;
	//	}
	
	//	/**
	//	 * test main
	//	 * 
	//	 * @param args
	//	 */
	//	public static void main(String[] args) {
	//		/*
	//		 * Initialize the Logger (Config file is located in folder 'xml' with name 'log4j.xml')
	//		 */
	//		org.apache.log4j.xml.DOMConfigurator.configureAndWatch("xml" + java.io.File.separator + "log4j.xml",
	//				60 * 1000);
	//		
	//		EsperController c = new EsperController(null);
	//		c.test();
	//	}
	
	/**
	 * Returns the newEventName of a given pattern
	 * 
	 * @param patternID id of the pattern
	 * @param selectFunctionNumber number of the select function which results are counted
	 * 
	 * @return the newEventName of the pattern
	 */
	public String getNewEventName(String patternID, int selectFunctionNumber) {
		//get all patterns from parser
		HashMap<String, APattern> patterns = parser.getPatterns();
		
		//search for pattern
		if (!this.parser.getPatterns().containsKey(patternID)) {
			logger.severe("pattern ID (" + patternID + ") not found");
			return null;
		}
		APattern pattern = patterns.get(patternID);
		
		//search for select function
		if (!(pattern.getSelectFunctions().size() > selectFunctionNumber)) {
			if (!(pattern.getSelectFunctions().size() >= 0)) {
				logger.severe("No select function and therefore no 'newEventName' defined in pattern '" + patternID
						+ "'. Can not use this pattern in a repetitive pattern.");
				return null;
			}
			logger.warning("The pattern with the id '" + patternID + "does not define at least "
					+ selectFunctionNumber + " selectfunctions. Using first select function instead.");
			
			//set number to 0
			selectFunctionNumber = 0;
		}
		
		//return newEventName
		return pattern.getSelectFunctions().get(selectFunctionNumber).getNewEventName();
	}
}
