/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml.esper;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.notification.SubscriptionManager;

import com.espertech.esper.client.UpdateListener;
import com.espertech.esper.event.EventBean;

import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.eml.pattern.SelFunction;
import de.ifgi.lehre.thesisEverding.eml.pattern.Statement;
import de.ifgi.lehre.thesisEverding.util.ThreadPool;

/**
 * listener for a single esper statement
 * 
 * @author Thomas Everding
 *
 */
public class StatementListener implements UpdateListener{
	
	private Statement statement;
	
	private EsperController controller;
	
	private boolean doOutput;
	
	private SubscriptionManager subMgr;
	
	/**
	 * 
	 * Constructor
	 *
	 * @param statement one {@link Statement}, used to configure this listener
	 * @param controller the esper controller with the esper engine
	 */
	public StatementListener(Statement statement, EsperController controller) {
		this.statement = statement;
		this.controller = controller;
		
		this.initialize();
	}
	
	public StatementListener(Statement statement, EsperController controller, SubscriptionManager sub) {
		this(statement, controller);
		this.subMgr = sub;
	}
	
	
	/**
	 * initializes the listener
	 */
	@SuppressWarnings("unchecked")
	private void initialize() {
		//check for output
//		if (this.statement.getSelectFunction().getOutputName().equals("")) {
//			this.doOutput = false;
//		}
//		else {
//			this.doOutput = true;
//		}
		this.doOutput = true;
		
		//register new event at esper engine
		SelFunction sel = this.statement.getSelectFunction();
		if (!sel.getNewEventName().equals("")) {
			String eventName = sel.getNewEventName();
			
			//common attributes
			HashMap<String, Object> eventProperties = new HashMap<String, Object>();
			eventProperties.put(MapEvent.START_KEY, Long.class);
			eventProperties.put(MapEvent.END_KEY, Long.class);
			eventProperties.put(MapEvent.CAUSALITY_KEY, Vector.class);
			
			//register every event attribute
			if (sel.isSingleValueOutput()) {
				for (String key : sel.getDataTypes().keySet()) {
					eventProperties.put(key, sel.getDataTypes().get(key));
				}
			}
			else {
				//nested properties
				HashMap<String, Object> nestedMap = new HashMap<String, Object>();
				for (String key : sel.getDataTypes().keySet()) {
					nestedMap.put(key, sel.getDataTypes().get(key));
				}
				
				if (nestedMap.get(MapEvent.VALUE_KEY) instanceof HashMap) {
					//get inner map
					nestedMap = (HashMap<String, Object>) nestedMap.get(MapEvent.VALUE_KEY);
				}
				
				//add nested properties
				for (String key : nestedMap.keySet()) {
					if (key.equals(MapEvent.START_KEY) || key.equals(MapEvent.END_KEY) || key.equals(MapEvent.CAUSALITY_KEY)) {
						//do nothing
					}
					else {
						eventProperties.put(key, nestedMap.get(key));
					}
					
				}
			}
			
			this.controller.registerEvent(eventName, eventProperties);
		}
	}

	public void update(EventBean[] newEvents, EventBean[] oldEvents) {
		/*
		 * new matches for the pattern received
		 * 
		 * handle every match
		 */
		if (newEvents != null) {
			for (EventBean newEvent : newEvents) {
				this.handleMatch(newEvent);
			}
		}
	}


	/**
	 * handles a single pattern match
	 * 
	 * @param newEvent the EventBean representing the match
	 */
	@SuppressWarnings("unchecked")
	private synchronized void handleMatch(EventBean newEvent) {
		UpdateHandlerThread handler = new UpdateHandlerThread(this, newEvent);
		
		//handle match in its own thread using a ThreadPool
		ThreadPool tp = ThreadPool.getInstance();
		tp.execute(handler);
	}
	
	
//	/**
//	 * Sends the received result to the controller for output
//	 * 
//	 * @param resultEvent the result to send
//	 */
//	public synchronized void doOutput(MapEvent resultEvent) {
//		String outputName = this.statement.getSelectFunction().getOutputName();
//		
//		//load output description
//		if (this.outDescription == null) {
//			if (this.getOutDescriptionPerformed) {
//				//output description not found
//				return;
//			}
//			
//			//try to find output description
//			this.outDescription = this.controller.getOutputDescription(outputName);
//			this.getOutDescriptionPerformed = true;
//			
//			if (this.outDescription == null) {
//				//not found
//				return;
//			}
//		}
//		
//		//send output (the whole event or only the value)
//		if (this.outDescription.getDataType().equals(SupportedDataTypes.EVENT)) {
//			//send event
//			this.controller.doOutput(outputName, resultEvent);
//		}
//		else {
//			//send only value
//			this.controller.doOutput(outputName, resultEvent.get(MapEvent.VALUE_KEY));
//		}
//	}
	
	/**
	 * Sends the received result to the controller for output
	 * 
	 * @param resultEvent the result to send
	 */
	@SuppressWarnings("unchecked")
	public synchronized void doOutput(MapEvent resultEvent) {
	    Map alert = (Map) resultEvent;
	    try {
	    	//get message and forward to SESSubscriptionManager
	    	NotificationMessage message = (NotificationMessage) alert.get(MapEvent.ORIGNIAL_MESSAGE_KEY);
			subMgr.publish(message);
		} catch (SoapFault e) {
			e.printStackTrace();
		}
	}
	



	/**
	 * @return the statement
	 */
	public Statement getStatement() {
		return this.statement;
	}


	/**
	 * @return the controller
	 */
	public EsperController getController() {
		return this.controller;
	}


	/**
	 * @return the doOutput
	 */
	public boolean isDoOutput() {
		return this.doOutput;
	}
}