/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml.esper;

import java.util.Date;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Logger;

import com.espertech.esper.event.*;

import de.ifgi.lehre.thesisEverding.eml.Constants;
import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.eml.pattern.Statement;


/**
 * Handles updates from a {@link StatementListener}.
 * 
 * @author Thomas Everding
 * 
 */
public class UpdateHandlerThread implements Runnable {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private EsperController controller;
	
	private Statement statement;
	
	private EventBean bean;
	
	private boolean doOutput;
	
	private StatementListener listener;
	
	
	/**
	 * 
	 * Constructor
	 * 
	 * @param listener listener that received the update
	 * @param bean the received update
	 */
	public UpdateHandlerThread(StatementListener listener, EventBean bean) {
		this.doOutput = listener.isDoOutput();
		this.controller = listener.getController();
		this.statement = listener.getStatement();
		this.bean = bean;
		this.listener = listener;
	}
	

	@SuppressWarnings("unchecked")
	public void run() {
		//build new event
		MapEvent event = null;
		
		if (bean instanceof MapEventBean) {
			MapEventBean selected = (MapEventBean) bean;
			String[] propertyNames = selected.getEventType().getPropertyNames();
			
			if (propertyNames.length > 1) {
				event = createEventFromComplexSelect(selected, propertyNames);
			}
			else if (propertyNames.length == 1){
				event = createEventFromSimpleSelect(event, propertyNames[0]);
			}
			else {
				event = createEventFromSimpleSelect(event);
			}
			
		}
		else {
			event = createEventFromSimpleSelect(event);
		}
		

		
		if (event == null) {
			//still nothing selected...
			logger.severe("no result generated from pattern update");
			return;
		} else {
			/*
			 * add original message
			 */
			Map alert = (Map) bean.getUnderlying();
			Object message = alert.get(MapEvent.ORIGNIAL_MESSAGE_KEY);
			if (message != null)
				event.put(MapEvent.ORIGNIAL_MESSAGE_KEY, message);
		}
		
		//create causality if wanted
		if (this.statement.getSelectFunction().getCreateCausality()) {
			MapEvent underlying = (MapEvent) bean.getUnderlying();
			
			Vector<MapEvent> underlyingCausality = (Vector<MapEvent>) underlying.get(MapEvent.CAUSALITY_KEY);
			
			//add causality of underlying event
			for (MapEvent e : underlyingCausality) {
				event.addCausalAncestor(e);
			}
			
			//add underlying event to causality
			event.addCausalAncestor(underlying);
		}
		
		//if output=true send event to output
		if (doOutput) {
			listener.doOutput(event);
		}
		
		//send event to esper engine for further processing
		this.controller.sendEvent(this.statement.getSelectFunction().getNewEventName(), event);
	}


	/**
	 * generates a new {@link MapEvent} from a selection with multiple values
	 * @param selected the selected {@link MapEventBean}
	 * @param propertyNames the available property names
	 * @return the new map event
	 */
	@SuppressWarnings("unchecked")
	private MapEvent createEventFromComplexSelect(MapEventBean selected, String[] propertyNames) {
		//event selected
		long start = Long.parseLong(selected.get(MapEvent.START_KEY).toString());
		long end = Long.parseLong(selected.get(MapEvent.END_KEY).toString());
		MapEvent event = new MapEvent(start, end);
		
		//copy content
		for (String key : propertyNames) {
			if (key.equals(MapEvent.START_KEY) || key.equals(MapEvent.END_KEY)) {
				//already copied
			}
			else if (key.equals(MapEvent.CAUSALITY_KEY)) {
				//copy causality
				Vector<MapEvent> causality = (Vector<MapEvent>) selected.get(key);
				
				for (MapEvent ancestor : causality) {
					event.addCausalAncestor(ancestor);
				}
			}
			else {
				event.put(key, selected.get(key));
			}
		}
		return event;
	}
	
	
	/**
	 * generates a new MapEvent from the selection of a single value
	 * @param event the selected event bean
	 * @return the new map event
	 */
	private MapEvent createEventFromSimpleSelect(MapEvent event) {
		return this.createEventFromSimpleSelect(event, MapEvent.VALUE_KEY);
	}


	/**
	 * generates a new MapEvent from the selection of a single value
	 * @param event the selected event bean
	 * @param propertyName the name of the only property
	 * @return the new map event
	 */
	private MapEvent createEventFromSimpleSelect(MapEvent event, String propertyName) {
		Date now = new Date();
		try {
			// no event selected, use only the property 'value'
			Object obj = bean.get(propertyName);
			
			if (obj == null) {
				return null;
			}
			
			//check for timer events
			if (obj.equals(Constants.TIMER_EVENT_VALUE)) {
				//timer event caught
				event = new MapEvent(now.getTime(), now.getTime());
				event.put(MapEvent.VALUE_KEY, now.getTime());
			}
			else {
				//handle object as usual
				event = new MapEvent(now.getTime(), now.getTime());
				
				event.put(MapEvent.VALUE_KEY, obj);
			}
		}
		catch (PropertyAccessException ex) {
			//no event property name value found
			logger.severe(ex.getMessage());
		}
		catch (Throwable t) {
			logger.severe(t.getMessage());
		}
		return event;
	}
}
