/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.eml.event;

import java.util.*;
import java.util.logging.Logger;

import org.apache.muse.ws.notification.NotificationMessage;


/**
 * Representation of events. Used in the esper engine for every event.
 * 
 * @author Thomas Everding, Matthes Rieke
 * 
 */
public class MapEvent implements Map<String, Object> {


	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	/**
	 * reserved key for the start timestamp
	 */
	public static final String START_KEY = "startTime";
	
	/**
	 * reserved key for the end timestamp
	 */
	public static final String END_KEY = "endTime";
	
	/**
	 * reserved key for the causal vector
	 */
	public static final String CAUSALITY_KEY = "causality";
	
	/**
	 * reserved key for the value of an event
	 */
	public static final String VALUE_KEY = "value";
	
	/**
	 * reserved key for the geometry of an event
	 */
	public static final String GEOMETRY_KEY = "geometry";
	
	/**
	 * reserved key for the sensorid of an event
	 */
	public static final String SENSORID_KEY = "sensorid";
	
	public static final String THIS_KEY = "this";

	public static final String ORIGNIAL_MESSAGE_KEY = "originalMessage";
	
	
	private HashMap<String, Object> map;
	
	private long start;
	
	private long end;
	
	private NotificationMessage originalMessage;
	
	/**
	 * 
	 * Constructor
	 * 
	 * @param startTime start time in ms
	 * @param endTime end time in ms (can be <= startTime for end = start)
	 *
	 */
	public MapEvent(long startTime, long endTime) {
		map = new HashMap<String, Object>();
		
		this.start = startTime;
		if (endTime > startTime) {
			this.end = endTime;
		}
		else {
			this.end = this.start;
		}
		
		this.initialize();
	}
	
	




	/**
	 * initializes the event map
	 */
	private void initialize() {
		//create causal vector
		Vector<MapEvent> causality = new Vector<MapEvent>();
		this.put(CAUSALITY_KEY,causality);
		
		//set start and end time
		this.put(START_KEY, this.start);
		this.put(END_KEY, this.end);
		
		//set this
		this.put(THIS_KEY, this);
	}
	
	
	/**
	 * adds a causal ancestor of this event
	 * 
	 * @param event the causal ancestor 
	 */
	@SuppressWarnings("unchecked")
	public void addCausalAncestor(MapEvent event) {
		((Vector<MapEvent>)this.get(CAUSALITY_KEY)).add(event);
	}
	

	public void clear() {
		map.clear();
		
		//initialize
		this.initialize();
	}
	

	public boolean containsKey(Object key) {
		return map.containsKey(key);
	}
	

	public boolean containsValue(Object value) {
		return map.containsValue(value);
	}
	

	public Set<java.util.Map.Entry<String, Object>> entrySet() {
		return map.entrySet();
	}
	

	public Object get(Object key) {
		return map.get(key);
	}
	

	public boolean isEmpty() {
		//a map event is never empty
		return false;
	}
	

	public Set<String> keySet() {
		return map.keySet();
	}
	

	public Object put(String key, Object value) {
		return map.put(key, value);
	}
	

	public void putAll(Map<? extends String, ? extends Object> m) {
		map.putAll(m);
	}
	

	public Object remove(Object key) {
		//causality and start/end time are not to be removed
		if (key.equals(START_KEY)
			|| key.equals(END_KEY)
			|| key.equals(CAUSALITY_KEY)
			|| key.equals(GEOMETRY_KEY)
			|| key.equals(SENSORID_KEY)
			|| key.equals(THIS_KEY)) {
			logger.severe("the key (" + key + ") can not be removed");
			return null;
		}
		
		return map.remove(key);
	}
	

	public int size() {
		return map.size();
	}
	

	public Collection<Object> values() {
		return map.values();
	}


	public NotificationMessage getOriginalMessage() {
		return originalMessage;
	}
	
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("MapEvent:\n");
		for (String key : map.keySet()) {
			if (!key.equals(MapEvent.THIS_KEY) && !key.equals(MapEvent.ORIGNIAL_MESSAGE_KEY)) {
				sb.append(key +"="+ map.get(key) +"\n");
			}
		}
		return sb.toString();
	}

	
}