/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.process;

import java.util.HashMap;
import java.util.Vector;
import java.util.logging.Logger;

import de.ifgi.lehre.thesisEverding.processCommunication.IInputConnector;
import de.ifgi.lehre.thesisEverding.processCommunication.IOutputConnector;
import de.ifgi.lehre.thesisEverding.util.DataBuffer;

/**
 * Abstract executor of a single process
 * 
 * @author Thomas Everding
 *
 */
public abstract class AProcessExecutor {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	protected HashMap<String, IInputConnector> inputConnectors = new HashMap<String, IInputConnector>();
	
	protected HashMap<String, Vector<IOutputConnector>> outputConnectors = new HashMap<String, Vector<IOutputConnector>>();
	
	protected boolean running = true;
	
	protected boolean computing = false;
	
	protected HashMap<String, Object> values;
	
	protected HashMap<String, Object> givenParameters;
	
	protected DataBuffer buffer;
	
	
	/**
	 * Starts the execution of the process
	 */
	protected abstract void startExecution();
	
	
	/**
	 * Starts the execution of the process
	 */
	public void start() {
		running = true;
		//start the execution in an new thread
		Thread t = new Thread(new Runnable() {
			
			public void run() {
				//start execution
				startExecution();				
			}
			
		});
		
		t.setDaemon(true);
		t.start();
	}
	
	
	/**
	 * stops the listeners and the execution
	 */
	public void stop() {	
		//stop execution
		running = false;
		
		//stop execution
		stopExecution();
	}
	
	
	/**
	 * Stops the execution of the process
	 */
	protected abstract void stopExecution();
	
	
	/**
	 * performs a single calculation of the 
	 * output of a data tuple
	 * 
	 * DO NOT CALL THIS METHOD DIRECTLY!
	 * USE <code>calculate()</code> INSTEAD!!
	 */
	protected abstract void doCalculation();
	
	
	/**
	 * performs a single calculation of the 
	 * output of a data tuple
	 * 
	 * calls  ends with Thread.yield();
	 */
	public synchronized void calculate() {
		if (!running) {
			//stop computing
			return;
		}
		computing = true;
		
		if (buffer != null) {
			//get values
			values = buffer.getData(inputConnectors.keySet());
			
			//check
			if (values == null) {
				//data not (yet) complete, abort
				computing = false;
				Thread.yield();
				return;
			}
			
			//copy given parameters to values
			for (String key : givenParameters.keySet()) {
				values.put(key, givenParameters.get(key));
			}
		}
		else {
			//buffer not available
			logger.warning("no buffer available");
		}
		
		//do calculation
		this.doCalculation();
		
		if (buffer != null && buffer.isUpdated()) {
			//new data available
			Thread.yield();
			calculate();
			return;
		}
		
		computing = false;
		
		//yield the thread
		Thread.yield();
	}
	
	
	/**
	 * adds an input connector to the list
	 * 
	 * @param inputName name of the connected input 
	 * @param connector the new input connector
	 */
	public void addInputConnector (String inputName, IInputConnector connector) {
		inputConnectors.put(inputName, connector);
	}
	
	
	/**
	 * adds an output connector to the list
	 * 
	 * @param outputName name of the connected output
	 * @param connector the new output connector
	 */
	public void addOutputConnector (String outputName, IOutputConnector connector) {
		if (outputConnectors.get(outputName) == null) {
			outputConnectors.put(outputName, new Vector<IOutputConnector>());
		}
		outputConnectors.get(outputName).add(connector);
	}


	/**
	 * @return <code>true</code> if the executor is computing
	 */
	public boolean isComputing() {
		return this.computing;
	}


	/**
	 * @param givenParameters the given parameters of the process
	 */
	public void setGivenParameters(HashMap<String, Object> givenParameters) {
		this.givenParameters = givenParameters;
	}
}
