/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.processCommunication.fifoQueue;

import java.util.logging.Logger;

/**
 * FIFO queue for the communication between
 * two processes. can be used as a stream
 * 
 * @author Thomas Everding
 *
 */
public class FIFOQueue {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private int elementCout;
	private int maxSize;
	
	private boolean isFull;
	private boolean isEmpty;
	
	private FIFOQueueElement first;
	private FIFOQueueElement last;
	
	
	/**
	 * 
	 * Constructor
	 * 
	 * Creates an unbounded FIFO queue.
	 * (maximum size = Integer.MAX_VALUE)
	 *
	 */
	public FIFOQueue() {
		this(Integer.MAX_VALUE);
	}
	
	
	/**
	 * 
	 * Constructor
	 * 
	 * Creates a bounded FIFO queue
	 *
	 * @param maxSize maximum element count
	 */
	public FIFOQueue(int maxSize) {
		this.maxSize = maxSize;
		if (maxSize <= 0) {
			logger.warning("Using a FIFO queue with size < 1 may result in a deadlock!");
		}
		
		this.elementCout = 0;
		this.isEmpty = true;
		this.isFull = this.elementCout >= this.maxSize;
	}
	
	
	/**
	 * Adds an item to the FIFO queue.
	 * If the queue is full this method will block.
	 * 
	 * @param item the new item
	 */
	public synchronized void addItem(Object item) {
		while (isFull) {
			//wait for free space
			try {
				wait();
			}
			catch (InterruptedException ex) {
			}
		}
		
		/*
		 * this code is only executed with exclusive 
		 * access and free space in the FIFO queue
		 */
		
		//add new element to the queue
		FIFOQueueElement newElem = new FIFOQueueElement(item);
		if (isEmpty) {
			//at first place if queue is empty
			this.first = newElem;
		}
		else {
			//else append at the end
			this.last.setNext(newElem);
		}
		this.last = newElem;
		
		//recalculate indices
		this.elementCout++;
		this.isEmpty = false;
		this.isFull = this.elementCout >= this.maxSize;
		
		//notify all waiting threads
		notifyAll();
	}
	
	
	/**
	 * If the queue is empty this method will block.
	 * 
	 * @return the first item in the queue.
	 */
	public synchronized Object getItem() {
		while (isEmpty) {
			//wait for an item
			try {
				wait();
			}
			catch (InterruptedException ex) {
			}
		}
		
		/*
		 * this code is only executed with exclusive
		 * access and at least one item in the FIFO
		 * queue
		 */
		
		//read item
		Object item = first.getItem();
		
		//remove first item from queue
		this.first = this.first.getNext();
		
		//recalculate indices
		this.elementCout--;
		this.isEmpty = this.elementCout <= 0;
		this.isFull = false;
		
		//notify all waiting threads and return item
		notifyAll();
		return item;
	}
}
