/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.sml;

import java.util.HashMap;
import java.util.Vector;
import java.util.logging.Logger;

import net.opengis.sensorML.x101.*;
import net.opengis.sensorML.x101.SensorMLDocument.SensorML;
import net.opengis.sensorML.x101.SensorMLDocument.SensorML.Member;
import net.opengis.swe.x101.DataComponentPropertyType;
import net.opengis.swex.x001.EventType;
import de.ifgi.lehre.thesisEverding.eml.constants.SupportedDataTypes;
import de.ifgi.lehre.thesisEverding.process.*;

/**
 * helps to parse a SensorML document
 * 
 * @author Thomas Everding
 *
 */
public class SensorMLParsingHelper {
	
	/*
	 * Logger instance for this class
	 */
	private static final Logger logger = Logger.getLogger(SensorMLParsingHelper.class.getName());
	
	private static final String VALUE_TAG = "<swe:value>";
	
	private static final String COMMENT_START_TAG = "<!--";
	
	private static final String COMMENT_END_TAG   = "-->";
	
	private SensorMLDocument smlDocument;
	
	private AbstractProcessType aProcess;
	
	private ESupportedProcessTypes processType = ESupportedProcessTypes.NotSupported;
	
	private boolean invalidDocument;
	
	private IoComponentPropertyType[] inputs;
	
	private IoComponentPropertyType[] outputs;
	
	private DataComponentPropertyType[] varyingParameters;
	
	private HashMap<String, Object> givenParameters;
	
	
	/**
	 * 
	 * Constructor
	 *
	 * @param smlDocument the SensorML document that shall be parsed
	 */
	public SensorMLParsingHelper(SensorMLDocument smlDocument) {
		this.smlDocument = smlDocument;
		this.invalidDocument = false;
	}
	
	
	/**
	 * parses the input descriptions from the SensorML document
	 * 
	 * @return the input descriptions
	 */
	public InputDescription[] parseInputDescriptions() {
		//check process
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		
		//check inputs
		if (inputs == null) {
			this.loadInputs();
		}
		
		//check for validity
		if (invalidDocument) {
			return new InputDescription[0];
		}
		
		InputDescription[] result = new InputDescription[inputs.length];
		
		//load every input
		for (int i = 0; i < inputs.length; i++) {
			result[i] = new InputDescription();
			loadDescription(inputs[i], result[i]);
		}
		
		return result;
	}
	
	
	/**
	 * parses the parameter descriptions from the SensorML document
	 * 
	 * @return the parameter descriptions
	 */
	public InputDescription[] parseParameterDescriptions() {
		//check process
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		
		//check parameters
		if (varyingParameters == null) {
			this.loadParameters();
		}
		
		//check for validity
		if (invalidDocument) {
			return new InputDescription[0];
		}
		
		//TODO
		InputDescription[] result = new InputDescription[varyingParameters.length];
		
		//load every parameter without value
		for (int i = 0; i < varyingParameters.length; i++) {
			result[i] = new InputDescription();
			loadDescription(varyingParameters[i], result[i]);
		}
		
		return result;
	}


	/**
	 * parses the output descriptions from the SensorML document
	 * 
	 * @return the output descriptions
	 */
	public OutputDescription[] parseOutputDescriptions() {
		//check process
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		
		//check outputs
		if (outputs == null) {
			this.loadOutputs();
		}
		
		//check for validity
		if (invalidDocument) {
			return new OutputDescription[0];
		}
		
		OutputDescription[] result = new OutputDescription[outputs.length];
		
		//load every output
		for (int i = 0; i < outputs.length; i++) {
			result[i] = new OutputDescription();
			loadDescription(outputs[i], result[i]);
		}
		
		return result;
	}
	
	
	/**
	 * loads the {@link AbstractProcessType} from the fist SensorML member
	 */
	private void loadAbstractProcess() {
		SensorML sml = smlDocument.getSensorML();
		Member[] membs = sml.getMemberArray();
		Member memb = membs[0];
		if (memb.isSetProcess()) {
			aProcess = memb.getProcess();
		}
		
		//check
		if (aProcess == null) {
			logger.severe("Invalid SensorML file!");
			invalidDocument = true;
			return;
		}
	}
	
	
	/**
	 * loads the process type from the abstract process
	 */
	private void loadProcessType() {
		if (aProcess == null) {
			this.loadAbstractProcess();
		}
		
		if (aProcess instanceof ProcessModelType) {
			//process model
			processType = ESupportedProcessTypes.ProcessModel;
		}
		else if (aProcess instanceof ProcessChainType) {
			//process chain
			processType = ESupportedProcessTypes.ProcessChain;
		}
		else {
			//not supported type
			logger.severe("the used process type is not supported");
			invalidDocument = true;
		}
	}
	
	
	/**
	 * loads the properties of a parameter description
	 * 
	 * @param dataComponent SensorML part with the description
	 * @param description description object of the connection
	 */
	public static void loadDescription(DataComponentPropertyType dataComponent, InputDescription description) {
		//set name
		description.setName(dataComponent.getName());
		
		//set underlying
		description.setParamUnderlying(dataComponent);
		
		//look for data type and uom
//		if (dataComponent.isSetAbstractDataArray1()) {
//			//array type
//			if (dataComponent.getAbstractDataArray1() instanceof DataArrayType) {
//				//data array type
//				description.setDataType(SupportedDataTypes.DATA_ARRAY);
//				
//				//check uom
//				description.setUOM(this.getUOMFromString(dataComponent.getAbstractDataArray1().toString()));
//			}
//			else {
//				//not supported array type
//				logger.error("this array type is not supported");
//			}
//		}
		if (dataComponent.isSetAbstractDataRecord()) {
			//record type
			if (dataComponent.getAbstractDataRecord() instanceof EventType) {
				//swe:Event type
				description.setDataType(SupportedDataTypes.EVENT);
			}
//			if (dataComponent.getAbstractDataRecord() instanceof SimpleDataRecordType) {
//				//simple data record type
//				description.setDataType(SupportedDataTypes.SIMPLE_DATA_RECORD);
//				
//				//check uom
//				description.setUOM(this.getUOMFromString(dataComponent.getAbstractDataRecord().toString()));
//			}
//			else {
//				//not supported record type
//				logger.error("this record type is not supported");
//			}
		}
		else if (dataComponent.isSetBoolean()) {
			//boolean type
			description.setDataType(SupportedDataTypes.BOOLEAN);
		}
		else if (dataComponent.isSetCategory()) {
			//category type
			description.setDataType(SupportedDataTypes.CATEGORY);
		}
		else if (dataComponent.isSetText()) {
			//text type
			description.setDataType(SupportedDataTypes.TEXT);
		}
		else if (dataComponent.isSetCount()) {
			//count type
			description.setDataType(SupportedDataTypes.COUNT);
		}
//		else if (dataComponent.isSetCountRange()) {
//			//count range type
//			description.setDataType(SupportedDataTypes.COUNT_RANGE);
//		}
		else if (dataComponent.isSetQuantity()) {
			//quantity type
			description.setDataType(SupportedDataTypes.QUANTITY);
			
			//check uom
			if (dataComponent.getQuantity().isSetUom()
				&& dataComponent.getQuantity().getUom().isSetCode()) {
				description.setUOM(dataComponent.getQuantity().getUom().getCode());
			}
		}
//		else if (dataComponent.isSetQuantityRange()) {
//			//quantity range type
//			description.setDataType(SupportedDataTypes.QUANTITY_RANGE);
//			
//			//check uom
//			if (dataComponent.getQuantityRange().isSetUom()
//				&& dataComponent.getQuantityRange().getUom().isSetCode()) {
//				description.setUOM(dataComponent.getQuantityRange().getUom().getCode());
//			}
//		}
		else if (dataComponent.isSetTime()) {
			//time type
			description.setDataType(SupportedDataTypes.TIME);
			
			//check uom
			if (dataComponent.getTime().isSetUom()
				&& dataComponent.getTime().getUom().isSetCode()) {
				description.setUOM(dataComponent.getTime().getUom().getCode());
			}
		}
//		else if (dataComponent.isSetTimeRange()) {
//			//time range type
//			description.setDataType(SupportedDataTypes.TIME_RANGE);
//			
//			//check uom
//			if (dataComponent.getTimeRange().isSetUom()
//				&& dataComponent.getTimeRange().getUom().isSetCode()) {
//				description.setUOM(dataComponent.getTimeRange().getUom().getCode());
//			}
//		}
		else {
			//no supported type
			logger.severe("data type not supported");
		}
	}
	
	
	/**
	 * loads the properties of a {@link IConnectionDescription}
	 * only for in- and outputs
	 * 
	 * @param ioComponent SensorML part with the description
	 * @param description description object of the connection
	 */
	public static void loadDescription(IoComponentPropertyType ioComponent, IConnectionDescription description) {		
		//set name
		description.setName(ioComponent.getName());
		
		//set underlying
		if (description instanceof InputDescription) {
			((InputDescription)description).setInputUnderlying(ioComponent);
		}
		
		//look for data type and uom
//		if (ioComponent.isSetAbstractDataArray1()) {
//			//array type
//			if (ioComponent.getAbstractDataArray1() instanceof DataArrayType) {
//				//data array type
//				description.setDataType(SupportedDataTypes.DATA_ARRAY);
//				
//				//check uom
//				description.setUOM(this.getUOMFromString(ioComponent.getAbstractDataArray1().toString()));
//			}
//			else {
//				//not supported array type
//				logger.error("this array type is not supported");
//			}
//		}
		if (ioComponent.isSetAbstractDataRecord()) {
			//record type
			if (ioComponent.getAbstractDataRecord() instanceof EventType) {
				//swe:Event type
				description.setDataType(SupportedDataTypes.EVENT);
			}
//			if (ioComponent.getAbstractDataRecord() instanceof SimpleDataRecordType) {
//				//simple data record type
//				description.setDataType(SupportedDataTypes.SIMPLE_DATA_RECORD);
//				
//				//check uom
//				description.setUOM(this.getUOMFromString(ioComponent.getAbstractDataRecord().toString()));
//			}
//			else {
//				//not supported record type
//				logger.error("this record type is not supported");
//			}
		}
		else if (ioComponent.isSetBoolean()) {
			//boolean type
			description.setDataType(SupportedDataTypes.BOOLEAN);
		}
		else if (ioComponent.isSetCategory()) {
			//category type
			description.setDataType(SupportedDataTypes.CATEGORY);
		}
		else if (ioComponent.isSetText()) {
			//text type
			description.setDataType(SupportedDataTypes.TEXT);
		}
		else if (ioComponent.isSetCount()) {
			//count type
			description.setDataType(SupportedDataTypes.COUNT);
		}
//		else if (ioComponent.isSetCountRange()) {
//			//count range type
//			description.setDataType(SupportedDataTypes.COUNT_RANGE);
//		}
		else if (ioComponent.isSetQuantity()) {
			//quantity type
			description.setDataType(SupportedDataTypes.QUANTITY);
			
			//check uom
			if (ioComponent.getQuantity().isSetUom()
				&& ioComponent.getQuantity().getUom().isSetCode()) {
				description.setUOM(ioComponent.getQuantity().getUom().getCode());
			}
		}
//		else if (ioComponent.isSetQuantityRange()) {
//			//quantity range type
//			description.setDataType(SupportedDataTypes.QUANTITY_RANGE);
//			
//			//check uom
//			if (ioComponent.getQuantityRange().isSetUom()
//				&& ioComponent.getQuantityRange().getUom().isSetCode()) {
//				description.setUOM(ioComponent.getQuantityRange().getUom().getCode());
//			}
//		}
		else if (ioComponent.isSetTime()) {
			//time type
			description.setDataType(SupportedDataTypes.TIME);
			
			//check uom
			if (ioComponent.getTime().isSetUom()
				&& ioComponent.getTime().getUom().isSetCode()) {
				description.setUOM(ioComponent.getTime().getUom().getCode());
			}
		}
//		else if (ioComponent.isSetTimeRange()) {
//			//time range type
//			description.setDataType(SupportedDataTypes.TIME_RANGE);
//			
//			//check uom
//			if (ioComponent.getTimeRange().isSetUom()
//				&& ioComponent.getTimeRange().getUom().isSetCode()) {
//				description.setUOM(ioComponent.getTimeRange().getUom().getCode());
//			}
//		}
		else {
			//no supported type
			logger.severe("data type not supported");
		}
	}
	

//	private String getUOMFromString(String str) {
//		String[] segments;
//		String uom = "";
//		
//		segments = str.split(UOM_TAG);
//		if (segments.length > 1) {
//			//uom described -> read
//			uom = segments[1].split("\"")[0];
//		}
//		
//		return uom;
//	}


	/**
	 * loads the inputs
	 */
	private void loadInputs() {
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		if (aProcess == null) {
			return;
		}
		if (processType == ESupportedProcessTypes.NotSupported) {
			return;
		}
		if (invalidDocument) {
			return;
		}
		
		//load inputs
		if (processType == ESupportedProcessTypes.ProcessChain) {
			ProcessChainType pc = (ProcessChainType) aProcess;
			if (!pc.isSetInputs() || !pc.getInputs().isSetInputList()) {
				//no input defined
				inputs = new IoComponentPropertyType[0];
			}
			else {
				//set inputs
				inputs = pc.getInputs().getInputList().getInputArray();
			}
		}
		else if (processType == ESupportedProcessTypes.ProcessModel) {
			ProcessModelType pm = (ProcessModelType) aProcess;
			if (!pm.isSetInputs() || !pm.getInputs().isSetInputList()) {
				//no inputs defined
				inputs = new IoComponentPropertyType[0];
			}
			else {
				//set inputs
				inputs = pm.getInputs().getInputList().getInputArray();
			}
		}
	}
	
	
	/**
	 * loads the parameters
	 */
	private void loadParameters() {
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		if (aProcess == null) {
			return;
		}
		if (processType == ESupportedProcessTypes.NotSupported) {
			return;
		}
		if (invalidDocument) {
			return;
		}
		
		//load parameters
		if (processType == ESupportedProcessTypes.ProcessChain) {
			ProcessChainType pc = (ProcessChainType) aProcess;
			if (!pc.isSetParameters() || !pc.getParameters().isSetParameterList()) {
				//no parameter defined
				varyingParameters = new DataComponentPropertyType[0];
			}
			else {
				//set parameters
				varyingParameters = pc.getParameters().getParameterList().getParameterArray();
			}
		}
		else if (processType == ESupportedProcessTypes.ProcessModel) {
			ProcessModelType pm = (ProcessModelType) aProcess;
			if (!pm.isSetParameters() || !pm.getParameters().isSetParameterList()) {
				//no parameters defined
				varyingParameters = new DataComponentPropertyType[0];
			}
			else {
				//set parameters
				varyingParameters = pm.getParameters().getParameterList().getParameterArray();
			}
		}
		
		//load parameters with given value
		this.givenParameters = new HashMap<String, Object>();
		String value;
		Vector<DataComponentPropertyType> vParams = new Vector<DataComponentPropertyType>();
		for (DataComponentPropertyType param : varyingParameters) {
			value = this.getValueFromString(param.toString());
			if (value != null) {
				//value for parameter given
				this.givenParameters.put(param.getName(), value);
			}
			else {
				//value not given
				vParams.add(param);
			}
		}
		
		//copy vParams to varyingParameters
		this.varyingParameters = new DataComponentPropertyType[vParams.size()];
		this.varyingParameters = vParams.toArray(varyingParameters);
	}
	
	
	/**
	 * extracts a value from an AnyData XML - String
	 * 
	 * @return the value as String or null if nothing found
	 */
	private String getValueFromString(String xmlString) {
		String[] segments;
		String[] subsegments;
		if (xmlString.contains(COMMENT_START_TAG)) {
			//remove comments
			segments = xmlString.split(COMMENT_START_TAG);
			
			xmlString = segments[0];
			
			for (int i = 1; i < segments.length; i++) {
				subsegments = segments[i].split(COMMENT_END_TAG);
				if (subsegments.length > 1) {
					xmlString += subsegments[1];
				}
			}
		}
		
		//get value
			
		segments = xmlString.split(VALUE_TAG);
		
		if (segments.length > 1) {
			//value given -> read
			return segments[1].split("<")[0];
		}
		
		//value not given
		return null;
	}
	

	/**
	 * loads the outputs
	 */
	private void loadOutputs() {
		if (processType == ESupportedProcessTypes.NotSupported) {
			this.loadProcessType();
		}
		if (aProcess == null) {
			return;
		}
		if (processType == ESupportedProcessTypes.NotSupported) {
			return;
		}
		
		//load outputs
		if (processType == ESupportedProcessTypes.ProcessChain) {
			ProcessChainType pc = (ProcessChainType) aProcess;
			if (!pc.isSetOutputs() || !pc.getOutputs().isSetOutputList()) {
				//no outputs
				outputs = new IoComponentPropertyType[0];
			}
			else {
				//set outputs
				outputs = pc.getOutputs().getOutputList().getOutputArray();
			}
		}
		else if (processType == ESupportedProcessTypes.ProcessModel) {
			ProcessModelType pm = (ProcessModelType) aProcess;
			if (!pm.isSetOutputs() || !pm.getOutputs().isSetOutputList()) {
				//no outputs defined
				outputs = new IoComponentPropertyType[0];
			}
			else {
				//set outputs
				outputs = pm.getOutputs().getOutputList().getOutputArray();
			}
		}
	}



	/**
	 * @return <code>true</code> if the document was found to be invalid
	 */
	public boolean isInvalidDocument() {
		return this.invalidDocument;
	}


	/**
	 * @return the input count
	 */
	public int getInputCount() {
		if (inputs == null) {
			this.loadInputs();
		}
		
		return this.inputs.length;
	}


	/**
	 * @return the output count
	 */
	public int getOutputCount() {
		if (outputs == null) {
			this.loadOutputs();
		}
		return this.outputs.length;
	}


	/**
	 * @return the abstract process
	 */
	public AbstractProcessType getAbstractProcess() {
		if (aProcess == null) {
			this.loadAbstractProcess();
		}
		return this.aProcess;
	}


	/**
	 * @return the given parameters 
	 * (these with values given in SensorML)
	 */
	public HashMap<String, Object> getGivenParameters() {
		if (this.givenParameters == null) {
			this.loadParameters();
		}
		
		return this.givenParameters;
	}

}
