/**
 * Part of the diploma thesis of Thomas Everding.
 * @author Thomas Everding
 */

package de.ifgi.lehre.thesisEverding.util;

import java.util.HashMap;
import java.util.Set;
import java.util.logging.Logger;

import de.ifgi.lehre.thesisEverding.process.AProcessExecutor;

/**
 * buffers the input data of a process for execution
 * 
 * @author Thomas Everding
 *
 */
public class DataBuffer {
	
	/*
	 * Logger instance for this class
	 */
	private Logger logger = Logger.getLogger(this.getClass().getName());
	
	private HashMap<String, Object> data;
	
	private AProcessExecutor executor;
	
	private boolean updated = false;
	
	
	/**
	 * 
	 * Constructor
	 *
	 * @param executor the executor for the data
	 */
	public DataBuffer(AProcessExecutor executor) {
		this.executor = executor;
		
		data = new HashMap<String, Object>();
	}
	
	
	/**
	 * Returns a copy of the data.
	 * If the data is not complete, <code>null</code> is returned.
	 * If the data is complete a copy is returned;
	 * 
	 * @param inputNames the names of the inputs
	 * @return the map containing the data
	 */
	public synchronized HashMap<String, Object> getData(Set<String> inputNames) {
		updated = false;
		
		HashMap<String, Object> dataCopy = new HashMap<String, Object>();
		for (String key : inputNames) {
			//check completeness
			if (!data.containsKey(key)) {
				//not complete
				logger.severe("'" + key + "' is not available");
				return null;
			}
			
			//copy
			dataCopy.put(key, data.get(key));
		}
		
		return dataCopy;
	}
	
	
	/**
	 * Adds a value to the data.
	 * If the data map already contains a values for
	 * a given input name, it will be replaced with the 
	 * new delivered value.
	 * If the executor is not computing, it will be started.
	 * 
	 * @param inputName name of the input
	 * @param value new data value
	 */
	public synchronized void setData (String inputName, Object value) {
		if (this.data.containsKey(inputName)) {
			//value already exists
			this.data.remove(inputName);
		}
		//add value
		this.data.put(inputName, value);
		updated = true;
		
		//start calculation
		if (!executor.isComputing()) {
			executor.calculate();
		}
	}
	
	
	/**
	 * 
	 * @return <code>true</code> if the data was 
	 * updated since the last reading
	 */
	public synchronized boolean isUpdated() {
		return updated;
	}
}
