/***************************************************************
Copyright 2007 52North Initiative for Geospatial Open Source Software GmbH

 Author: jtheuer, University of Muenster

 Contact: Andreas Wytzisk, 
 52North Initiative for Geospatial Open Source SoftwareGmbH, 
 Martin-Luther-King-Weg 24,
 48155 Muenster, Germany, 
 info@52north.org

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2 as published by the Free Software Foundation.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; even without the implied WARRANTY OF
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program (see gnu-gpl v2.txt). If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA or visit the Free
 Software Foundations web page, http://www.fsf.org.

 ***************************************************************/

package org.n52.swe.sas.muse.common;

import java.io.InputStream;
import java.util.Collection;
import java.util.logging.Logger;

import net.opengis.ses.x100.CapabilitiesDocument;
import net.opengis.ses.x100.DescribeSensorDocument;
import net.opengis.ses.x100.ContentsDocument.Contents;
import net.opengis.ses.x100.ContentsDocument.Contents.RegisteredSensors;
import net.opengis.ses.x100.DescribeSensorDocument.DescribeSensor;
import net.opengis.ses.x100.GetCapabilitiesDocument.GetCapabilities;

import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.resource.impl.AbstractWsResourceCapability;
import org.apache.muse.ws.resource.sg.impl.AddHandler;
import org.apache.xmlbeans.XmlObject;
import org.n52.swe.sas.bindings.InvalidXMLContentException;
import org.n52.swe.sas.bindings.XMLBeansParser;
import org.n52.swe.sas.muse.wsbr.*;
import org.w3c.dom.*;

/**
 * @author Jan Torben Heuer <jan.heuer@uni-muenster.de>
 * 
 * This capability informs about what this service provides. Available sensors
 * are listed und can be futher described.
 */
public class Capabilites extends AbstractWsResourceCapability implements ICapabilites {
	/* automatically generated Logger */@SuppressWarnings("unused")
	private static final Logger LOGGER = Logger.getLogger(Capabilites.class.getName());
	private static final String CAPABILITIESFILE = "capabilites_template.xml";
	private RegisterPublisher registerpublisher;
	private CapabilitiesDocument capabilities;

	@Override
	public void initialize() throws SoapFault {
		super.initialize();
		
		setMessageHandler(new SimpleHandler("http://www.opengis.net/ses/GetCapabilitiesRequest",GetCapabilities.type.getName(),this,"getCapabilities"));
		setMessageHandler(new SimpleHandler("http://www.opengis.net/ses/DescribeSensorRequest",DescribeSensor.type.getName(),this,"describeSensor"));
		
		this.registerpublisher = getRegisterPublisher();
		this.capabilities = getCapabilityTemplate();
	}

	/**
	 * Checks if the capabilities template exists
	 * 
	 * @return the template or a new instance of CapabilitesDocument
	 * @throws
	 */
	private CapabilitiesDocument getCapabilityTemplate() throws SoapFault {
		CapabilitiesDocument capabilites;

		InputStream stream = getClass().getResourceAsStream(CAPABILITIESFILE);

		/* check if a template exists */
		if (stream == null) {
			capabilites = CapabilitiesDocument.Factory.newInstance();
			capabilites.addNewCapabilities();
		} else {
			try {
				capabilites = (CapabilitiesDocument) XMLBeansParser.parse(stream);
			} catch (InvalidXMLContentException e) {
				throw new SoapFault("The capabilites template is invalid! The error is:" + e.getMessage());
			} catch (ClassCastException e) {
				throw new SoapFault("The capabilites template is invalid! It doesn't seem to be a valid Capabilites document.");
			}
		}

		capabilites.getCapabilities().setVersion("3.0.0");
		return capabilites;
	}

	/**
	 * looks up the <em>first</em> {@link RegisterPublisher} instance
	 * 
	 * @return its reference
	 * @throws SoapFault
	 */
	protected RegisterPublisher getRegisterPublisher() throws SoapFault {
		try {
			final RegisterPublisher wsn = (RegisterPublisher) getResource().getCapability(WsbrConstants.NAMESPACE_URI);

			if (wsn == null) {
				String all = getResource().getCapabilityURIs().toString();
				throw new SoapFault("Cannot find the RegisterPublisher instance. Known instances are: " + all);
			} else {
				return wsn;
			}
		} catch (ClassCastException e) {
			throw new SoapFault("Internal Server error. Expected " + WsbrConstants.NAMESPACE_URI + " to be handled by "
					+ RegisterPublisher.class.getCanonicalName());
		}
	}

	public Element describeSensorReflect(Object[] o) throws Exception {
		return describeSensor((Element) o[0]);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.n52.swe.sas.muse.common.ICapabilites#describeSensor(org.w3c.dom.Element)
	 */
	public Element describeSensor(Element sensor_element) throws Exception {

		XmlObject xmlobject = XmlObject.Factory.parse(sensor_element);
		String sensorID;
		
		if(xmlobject instanceof DescribeSensorDocument) {
			sensorID=((DescribeSensorDocument)xmlobject).getDescribeSensor().getSensorID();
		} else {
			throw new SoapFault(sensor_element.toString() + " is not a valid sensor identifier");
		}
		
		
		Collection<PublisherEndpoint> endpoints = registerpublisher.getPublisherEndpoints();

		/*
		 * iterate over all endpoints. Could be made faster with either caching
		 * or hashes.
		 */
		for (PublisherEndpoint publisherEndpoint : endpoints) {
			if (publisherEndpoint.getSensorId().equals(sensorID)) {
				return publisherEndpoint.getSensorML();
			}
		}

		throw new SoapFault("Sensor " + sensor_element + " does not exist. Please check the capabilites or report an error");
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see org.n52.swe.sas.muse.common.ICapabilites#getCapabilities()
	 */
	public Element getCapabilities(Element xml) throws SoapFault {
		LOGGER.info("creating capabilites");
		CapabilitiesDocument temp = (CapabilitiesDocument) capabilities.copy();

		/* add content section if missing */
		Contents content = temp.getCapabilities().getContents();
		if (content == null) {
			content = temp.getCapabilities().addNewContents();
		}

		/* read available sensors */
		Collection<PublisherEndpoint> endpoints = registerpublisher.getPublisherEndpoints();
		if (endpoints.size() > 0) {
			RegisteredSensors sensors = content.addNewRegisteredSensors();

			for (PublisherEndpoint publisherEndpoint : endpoints) {
				String sensor = publisherEndpoint.getSensorId();
				sensors.addSensorID(sensor);
			}
		}
		
		try {
			
			XMLBeansParser.validate(temp);
			LOGGER.info("capabilites are valid!");
		} catch (InvalidXMLContentException e) {
			throw new SoapFault("Internal server error: capabilities are not valid: " + e.getMessage());
		}

		
		
		Node node = temp.getCapabilities().getDomNode();
		if(node instanceof Element) {
			LOGGER.info("returning capabilites");
			return (Element) temp.getCapabilities().getDomNode();
		} else {
			throw new AssertionError("Expected " + node.toString() + " to be an org.w3c.dom.Element, but it is " + node.getClass().getCanonicalName());
		}
	}
}
