package org.n52.swe.sas.muse.filter;

import java.io.IOException;

import javax.xml.namespace.QName;

import net.opengis.eml.x001.EMLDocument;
import net.opengis.eml.x001.SimplePatternType;
import net.opengis.eml.x001.EMLDocument.EML;
import net.opengis.gml.MeasureType;
import net.opengis.ogc.*;

import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.notification.*;
import org.n52.swe.sas.muse.unitconversion.IUnitConverter;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;


/**
 * This is the wrapper class for Level-2/3-Filtering.
 * 
 * @author Matthes Rieke <m.rieke@uni-muenster.de>
 *
 */
public class SESConstraintFilter implements Filter {

	private static final QName AND_QNAME = new QName("http://www.opengis.net/ogc", "And");
	
	private static final QName OR_QNAME = new QName("http://www.opengis.net/ogc", "Or");
	
	private static final QName NOT_QNAME = new QName("http://www.opengis.net/ogc", "Not");
	
	
	private EMLDocument emlXml;
	private FilterDocument filterXml;

	/**
	 * helper variable for unitconversion
	 */
	private String lastUrn;

	/**
	 * helper variable for unitconversion
	 */
	private String lastUom;

	public SESConstraintFilter(EMLDocument eml) {
		this.emlXml = eml;
	}

	public SESConstraintFilter(FilterDocument filter) {
		this.filterXml = filter;
	}


	public SESConstraintFilter(EML eml) {
		emlXml = EMLDocument.Factory.newInstance();
		emlXml.setEML(eml);
	}


	public EML getEml() {
		return emlXml.getEML();
	}

	public boolean accepts(NotificationMessage message) {
		return false;
	}

	public Element toXML() {
		return toXML(XmlUtils.EMPTY_DOC);
	}

	public Element toXML(Document doc) {
		Element filter = XmlUtils.createElement(doc, WsnConstants.FILTER_QNAME);

		//hack to get the <Filter> Element of OGCFilter
		String xmlText = "<MessageContent>\n";
		if (filterXml != null)
			xmlText += filterXml.toString();
		else if (emlXml != null)
			xmlText += emlXml.toString();
		else
			return null;
		xmlText += "\n</MessageContent>";

		Element message = null;
		try {
			Element node = XmlUtils.createDocument(xmlText).getDocumentElement();
			message = XmlUtils.createElement(doc, WsnConstants.MESSAGE_CONTENT_QNAME,
					node);
			message.setAttribute(WsnConstants.DIALECT, "http://52north.org/ses/contentfilter");
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}

		filter.appendChild(message);

		return filter;
	}


	/**
	 * replaces phenomenon Strings containing ":" to "__" and
	 * converts any quantity units to its base units.
	 * @throws Exception 
	 */
	public void replacePhenomenonStringsAndConvertUnits(IUnitConverter converter) throws Exception {
		if (emlXml != null) {
			FilterType filter = null;
			SimplePatternType[] patterns = emlXml.getEML().getSimplePatterns().getSimplePatternArray();
			for (SimplePatternType spt : patterns) {
				if (spt.isSetGuard()) {
					filter = spt.getGuard().getFilter();
					
					//TODO is stringreplace needed? or is it all done by the eml-parser?
					replaceForFilter(filter, converter);
				}

			}

		}

	}

	/**
	 * helpermethod for replacePhenomenonStringsAndConvertUnits().
	 * @throws Exception 
	 */
	private void replaceForFilter(FilterType filter, IUnitConverter converter) throws Exception {
		if (filter.isSetLogicOps()) {
			replaceForLogicOp(filter.getLogicOps(), converter);
		}
		
		if (filter.isSetComparisonOps()) {
			ComparisonOpsType cOps = filter.getComparisonOps();
			replaceForComparisonOp(cOps, converter);	
		}
		
	}
	
	/**
	 * helpermethod for replacePhenomenonStringsAndConvertUnits().
	 * @throws Exception 
	 */
	private void replaceForLogicOp(LogicOpsType logicOps, IUnitConverter converter) throws Exception {
		QName loQName = logicOps.newCursor().getName();
		
		/*
		 * check Not
		 */
		if (NOT_QNAME.equals(loQName)) {
			//create new NotFilter
			UnaryLogicOpType unaryOp = (UnaryLogicOpType) logicOps;
			if (unaryOp.isSetComparisonOps()) {
				replaceForComparisonOp(unaryOp.getComparisonOps(), converter);
			}
			if (unaryOp.isSetLogicOps()) {
				//rekursion
				replaceForLogicOp(unaryOp.getLogicOps(), converter);
			}
		}
		
		/*
		 * binary operators
		 */
		if (AND_QNAME.equals(loQName) || OR_QNAME.equals(loQName)) {
			BinaryLogicOpType binaryOp = (BinaryLogicOpType) logicOps;
			
			ComparisonOpsType[] array = binaryOp.getComparisonOpsArray();
			for (ComparisonOpsType cOps : array) {
				replaceForComparisonOp(cOps, converter);
			}
			
			LogicOpsType[] array2 = binaryOp.getLogicOpsArray();
			for (LogicOpsType logicOp : array2) {
				replaceForLogicOp(logicOp, converter);
			}
		}
				
	}
	
	/**
	 * helpermethod for replacePhenomenonStringsAndConvertUnits().
	 * @throws Exception 
	 */
	private void replaceForComparisonOp(ComparisonOpsType cOps, IUnitConverter converter) throws Exception {
		if (cOps instanceof PropertyIsBetweenType) {
			PropertyIsBetweenType pibt = (PropertyIsBetweenType) cOps;
			
			ExpressionType[] exps = new ExpressionType[3];
			exps[0] = pibt.getExpression();
			exps[1] = pibt.getLowerBoundary().getExpression();
			exps[2] = pibt.getUpperBoundary().getExpression();
			
			for (ExpressionType et : exps) {
				replaceForExpression(et, converter);
			}
			
		} else if (cOps instanceof BinaryComparisonOpType) {

			BinaryComparisonOpType bcot = (BinaryComparisonOpType) cOps;
			ExpressionType[] exps = bcot.getExpressionArray();
			for (ExpressionType et : exps) {
				replaceForExpression(et, converter);
			}
		} else if (cOps instanceof PropertyIsNullType) {
			PropertyIsNullType pint = (PropertyIsNullType) cOps;
			PropertyNameType pnt = pint.getPropertyName();
			replaceForExpression(pnt, null);

		} else if (cOps instanceof PropertyIsLikeType) {
			PropertyIsLikeType pilt = (PropertyIsLikeType) cOps;

			LiteralType lt = pilt.getLiteral();
			replaceForExpression(lt, converter);
								
			PropertyNameType pnt = pilt.getPropertyName();
			replaceForExpression(pnt, null);
		}
	}

	/**
	 * helpermethod for replacePhenomenonStringsAndConvertUnits().
	 * @throws Exception 
	 */
	private void replaceForExpression(ExpressionType et, IUnitConverter converter) throws Exception {
		if (et instanceof LiteralType) {
			LiteralType lt = (LiteralType) et;
			MeasureType quantity = lt.getQuantity();
			double value = quantity.getDoubleValue();
			String uom = quantity.getUom();
			
			if (lastUrn != null) {
				if (!converter.isCompatibleWithPhenomenon(uom, lastUrn)) {
					throw new Exception ("Unit is not compatible with previously registered" +
							"unit for this phenomenon");
				}
				lastUrn = null;
			}
			
			Object[] result = converter.convert(uom, value);
			quantity.setDoubleValue((Double) result[1]);
			quantity.setUom(result[0].toString());
			
			lastUom = uom;
		} else if (et instanceof PropertyNameType) {
			PropertyNameType pnt = (PropertyNameType) et;
			Element elem = (Element) pnt.getDomNode();
			String urn = XmlUtils.toString(elem.getFirstChild());
			
			if (lastUrn != null) {
				if (!converter.isCompatibleWithPhenomenon(lastUom, urn)) {
					throw new Exception ("Unit is not compatible with previously registered" +
							"unit for this phenomenon");
				}
				lastUom = null;
			}
			
			urn = urn.replaceAll(":", "__").replaceAll("\\.", "_");
			XmlUtils.setElementText(elem, urn);
			
			lastUrn = urn;
		}

	}

	
	
	@Override
	public String toString() {
		return XmlUtils.toString(toXML(), false);
	}

	public FilterType getFilter() {
		if (filterXml != null)
			return filterXml.getFilter();
		return null;
	}

}
