package org.n52.swe.sas.muse.filter.engine;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import net.opengis.eml.x001.EMLDocument.EML;
import net.opengis.ogc.FilterType;
import net.opengis.om.x10.ObservationDocument;

import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.Filter;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.notification.SubscriptionManager;
import org.apache.xmlbeans.XmlException;
import org.n52.swe.sas.muse.filter.SESConstraintFilter;
import org.n52.swe.sas.muse.filter.SESFilterCollection;
import org.n52.swe.sas.muse.filter.StaticEMLDocument;
import org.n52.swe.sas.muse.parser.OMParser;
import org.n52.swe.sas.muse.unitconversion.IUnitConverter;
import org.n52.swe.sas.muse.util.AdvancedResourceManagerListener;
import org.n52.swe.sas.muse.wsbr.PublisherEndpoint;
import org.n52.swe.sas.muse.wsn.SESSubscriptionManager;
import org.w3c.dom.Element;

import com.espertech.esper.client.EPServiceProvider;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKTReader;

import de.ifgi.lehre.thesisEverding.eml.esper.EsperController;
import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;
import de.ifgi.lehre.thesisEverding.eml.pattern.APattern;
import de.ifgi.lehre.thesisEverding.eml.pattern.PatternSimple;


/**
 * 
 * @author Matthes Rieke <m.rieke@uni-muenster.de>
 *
 */
public class FilterEngine implements IFilterEngine, AdvancedResourceManagerListener<PublisherEndpoint> {

	private Map<EPServiceProvider, String> esperSPs;
	private IUnitConverter unitConverter;
	private Random rand;
	private Logger logger;


	public FilterEngine(IUnitConverter converter, Logger logger) {
		this.logger = logger;
		unitConverter = converter;
		rand = new Random();
		//multiple runtimes needed for pattern management
		esperSPs = new HashMap<EPServiceProvider, String>();
	}

	public void filter(NotificationMessage message) {

		List<MapEvent> alerts = new ArrayList<MapEvent>();
		try {
			alerts = parseMessage(message);
		} catch (XmlException e) {
			getLog().log(Level.WARNING, e.getMessage(), e);
		} catch (Exception e) {
			getLog().log(Level.WARNING, e.getMessage(), e);
		}
		
		/*
		 * add the original Message
		 */
		if (alerts != null) {
			for (MapEvent mapEvent : alerts) {
				mapEvent.put(MapEvent.ORIGNIAL_MESSAGE_KEY, message);
			}
		}

		if (esperSPs != null) {
			for (MapEvent alert : alerts) {
				getLog().info("#!!# MapEvent = "+ alert);

				for (EPServiceProvider epsp : esperSPs.keySet()) {
					epsp.getEPRuntime().sendEvent(alert, esperSPs.get(epsp));
				}
			}	
		}

	}
	
	/**
	 * Parses O&M document and returns instances of MapEvent.
	 * @throws Exception 
	 */
	private List<MapEvent> parseMessage(NotificationMessage message) throws Exception {
		QName omQName = new QName("http://www.opengis.net/om/1.0", "Observation");
		Element content = message.getMessageContent(omQName);
		
		if (content != null) {
			ObservationDocument oDoc = ObservationDocument.Factory.parse(content);
			OMParser parser = new OMParser(oDoc, getLog());
			return parser.parseOM(unitConverter);
		}
		else {
			getLog().warning("could not handle data - There was no om:Observation element.");
		}
		
		return null;
	}



	/**
	 * Creates dummy alerts.
	 */
	@SuppressWarnings("unused")
	private List<MapEvent> getAlertsFromMessage(NotificationMessage message) {

		//dummy alerts
		String uom1 = "cel(1 K)";
		String phen1 = "urn__ogc__temperature";
		double val1 = rand.nextDouble()*5 + 20;

		Object[] result = null;
		try {
			result = unitConverter.convert(uom1, val1);
		} catch (Exception e) {
			e.printStackTrace();
		}

		List<MapEvent> alerts = new ArrayList<MapEvent>();

		String wktString = "POINT(15.1 15.2)";

		WKTReader wktReader = new WKTReader();

		Geometry geom = null;
		try {
			geom = wktReader.read(wktString);
		} catch (ParseException e) {
			e.printStackTrace();
		}

		MapEvent event = new MapEvent(System.currentTimeMillis() - 10000, System.currentTimeMillis());
		event.put(MapEvent.ORIGNIAL_MESSAGE_KEY, message);
		event.put(phen1, result[1]);
		event.put(MapEvent.SENSORID_KEY, "sensor-1");
		event.put(MapEvent.GEOMETRY_KEY, geom);
		
		alerts.add(event);

		return alerts;
	}
	
	
	public static void main(String[] args) {
		new FilterEngine(null, null).testRegisterFilter();
	}
	
	private void testRegisterFilter() {
		String file = "/subscribe_one_ogcfilter.xml";
		InputStream stream = getClass().getResourceAsStream(file);
		EML eml = null;
		try {
			eml = StaticEMLDocument.getWrapperDocument(FilterType.Factory.parse(stream));
		} catch (IOException e) {
			e.printStackTrace();
		} catch (XmlException e) {
			e.printStackTrace();
		}
		
		SESConstraintFilter filter = new SESConstraintFilter(eml);
		
		SESSubscriptionManager mgr = new SESSubscriptionManager();
		mgr.setFilter(filter);
		try {
			registerFilter(mgr);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@SuppressWarnings("unchecked")
	public void registerFilter(SubscriptionManager subMgr) throws Exception {

		SESSubscriptionManager sesSub;

		if (subMgr instanceof SESSubscriptionManager) {
			sesSub = (SESSubscriptionManager) subMgr;
		} else {
			throw new Exception("SESSubscriptionManager needed for registering a Filter.");
		}

		if (!sesSub.isHasConstraintFilter()) {
			throw new Exception("FilterEngine needs a SESConstraintFilter.");
		}

		Filter originalFilter = subMgr.getFilter();
		SESConstraintFilter filter;

		if (originalFilter instanceof SESConstraintFilter) {
			filter = (SESConstraintFilter) originalFilter;
		} else if (originalFilter instanceof SESFilterCollection) {
			SESFilterCollection filterColl = (SESFilterCollection) originalFilter;
			filter = filterColl.getConstraintFilter();
		} else {
			throw new Exception("FilterEngine needs SESConstraintFilter.");
		}
		

		/*
		 * do unit conversion,
		 * this also checks IUnitConverter.isCompatibleWithPhenomenon()
		 * for all phenomenons
		 */
		filter.replacePhenomenonStringsAndConvertUnits(unitConverter);
		
		/*
		 * parse the EML and get the patterns
		 */
		EsperController controller = new EsperController(subMgr);
		controller.initialize(filter.getEml(), null, null, null);

		Map<String, APattern> patternMap = controller.getParser().getPatterns();
		Map<String, APattern> simplePatterns = new HashMap<String, APattern>();
		
		for (String key : patternMap.keySet()) {
			APattern value = patternMap.get(key);
			if (value instanceof PatternSimple) {
				simplePatterns.put(key, value);
			}
		}

		//get streamName from EML
		String streamName = "";
		for (String key : simplePatterns.keySet()) {
			PatternSimple val = (PatternSimple) simplePatterns.get(key);
			if (!streamName.equals("") && !streamName.equals(val.getInputName())) {
				getLog().severe("Multiple stream names for one EML document!");
			}
			streamName = val.getInputName();
		}
		
		esperSPs.put(controller.getEpService(), streamName);
	}
	
	
	
//	@SuppressWarnings("unchecked")
//	public void registerFilter(SubscriptionManager subMgr) throws Exception {
//
//		SESSubscriptionManager sesSub;
//
//		if (subMgr instanceof SESSubscriptionManager) {
//			sesSub = (SESSubscriptionManager) subMgr;
//		} else {
//			throw new Exception("SESSubscriptionManager needed for registering a Filter.");
//		}
//
//		if (!sesSub.isHasConstraintFilter()) {
//			throw new Exception("FilterEngine needs a SESConstraintFilter.");
//		}
//
//		Filter originalFilter = subMgr.getFilter();
//		SESConstraintFilter filter;
//
//		if (originalFilter instanceof SESConstraintFilter) {
//			filter = (SESConstraintFilter) originalFilter;
//		} else if (originalFilter instanceof SESFilterCollection) {
//			SESFilterCollection filterColl = (SESFilterCollection) originalFilter;
//			filter = filterColl.getConstraintFilter();
//		} else {
//			throw new Exception("FilterEngine needs SESConstraintFilter.");
//		}
//		
//
//		/*
//		 * do unit conversion,
//		 * this also checks IUnitConverter.isCompatibleWithPhenomenon()
//		 * for all phenomenons
//		 */
//		filter.replacePhenomenonStringsAndConvertUnits(unitConverter);
//		
//		/*
//		 * parse the EML and get the patterns
//		 */
//		EMLParser eParser = new EMLParser(null);
//		eParser.parseEML(filter.getEml());
//
//		HashMap<String, APattern> patternMap = eParser.getPatterns();
//		String stmtString = "";
//		
//		/*
//		 * TODO Level-3-Filtering (Thomas?)
//		 */
//		APattern val = null;
//		for (String key : patternMap.keySet()) {
//			val = patternMap.get(key);
//			Statement[] stmts = val.createEsperStatements(eParser);
//
//			//TODO multiple SelectFunctions. Level-3-Filtering
//			stmtString = stmts[0].getStatement();
//		}
//
//		//register Map as event type with registered phenomenons/types
//		Map eventProperties = new HashMap<String, Object>();
//		eventProperties.put(MapEvent.START_KEY, Long.class);
//		eventProperties.put(MapEvent.END_KEY, Long.class);
//		eventProperties.put(MapEvent.CAUSALITY_KEY, Vector.class);
//		eventProperties.put(MapEvent.GEOMETRY_KEY, Geometry.class);
//		eventProperties.put(MapEvent.SENSORID_KEY, String.class);
//		eventProperties.put(MapEvent.THIS_KEY, Map.class);
//
//		/*
//		 * Get data types for phenomenons.
//		 */
//		DataTypesMap dtm = DataTypesMap.getInstance();
//
//		for (Object key : val.getPropertyNames()) {
//			eventProperties.put(key.toString(), dtm.getDataType(key.toString()));
//		}
//
//		Configuration config = new Configuration();
//		
//		//TODO get streamName from EML
//		String streamName = "sensorStream";
//		config.addEventTypeAliasNestable(streamName, eventProperties);
//		config.addImport("de.ifgi.lehre.thesisEverding.eml.esper.customFunctions.*");
//
//		EPServiceProvider esperService = EPServiceProviderManager.getProvider(
//				"ses:id:"+ subMgr.hashCode(), config);
//
//		//register statement
//		createStatement(stmtString, subMgr, esperService);
//
//		getLog().info("Registered new Statement: "+ stmtString);
//		
//		esperSPs.put(esperService, streamName);
//	}
//
//	private EPStatement createStatement(String stmtString, SubscriptionManager subMgr,
//			EPServiceProvider esperService) {
//
//		EPStatement stmt = esperService.getEPAdministrator().createEPL(stmtString);
//		stmt.addListener(new BasicListener(subMgr));
//		return stmt;
//	}

	
	
	private Logger getLog() {
		return logger;
	}

	/* (non-Javadoc)
	 * @see org.n52.swe.sas.muse.util.AdvancedResourceManagerListener#resourceAdded(org.apache.muse.ws.addressing.EndpointReference, java.lang.Object)
	 */
	public void resourceAdded(EndpointReference epr, PublisherEndpoint resource) throws SoapFault {}

	/* (non-Javadoc)
	 * @see org.n52.swe.sas.muse.util.AdvancedResourceManagerListener#resourceRemoved(org.apache.muse.ws.addressing.EndpointReference, java.lang.Object)
	 */
	public void resourceRemoved(EndpointReference epr, PublisherEndpoint resource) throws SoapFault {}

}
