package org.n52.swe.sas.muse.parser;

import java.math.BigInteger;
import java.text.ParseException;
import java.util.*;
import java.util.logging.Logger;

import net.opengis.gml.*;
import net.opengis.gml.EnvelopeType;
import net.opengis.gml.TimePositionType;
import net.opengis.om.x10.*;
import net.opengis.om.x10.ObservationDocument;
import net.opengis.om.x10.ObservationType;
import net.opengis.sampling.x10.*;
import net.opengis.swe.x101.*;
import net.opengis.swe.x101.TextBlockDocument.TextBlock;

import org.apache.muse.util.xml.XmlUtils;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.impl.values.XmlAnyTypeImpl;
import org.joda.time.DateTime;
import org.n52.swe.sas.muse.unitconversion.IUnitConverter;
import org.w3c.dom.Element;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.io.WKTReader;

import de.ifgi.lehre.thesisEverding.eml.event.MapEvent;

public class OMParser {

	private ObservationDocument document;
	private final Logger logger;

	public OMParser(ObservationDocument doc, Logger log) {
		this.document = doc;
		this.logger = log;
	}

	/**
	 * @param converter
	 * @return
	 * @throws Exception
	 */
	public List<MapEvent> parseOM(IUnitConverter converter) throws Exception {
		ArrayList<MapEvent> events = new ArrayList<MapEvent>();
		
		ObservationType obs = document.getObservation();
		List<PhenomenonUomModel> pums = new LinkedList<PhenomenonUomModel>();

		
		boolean hasTimestamp = false;
		
		/*
		 * Parse om:result. Use this phenomenons (and timestamps)
		 */
		XmlObject result = obs.getResult();
		
		if (result instanceof XmlAnyTypeImpl) {
			XmlAnyTypeImpl anyTypeDoc = (XmlAnyTypeImpl) result;

			DataArrayDocument dat = DataArrayDocument.Factory.parse(anyTypeDoc.toString());

			if (dat != null) {
				DataArrayType array = dat.getDataArray1();
				BigInteger elemCount = array.getElementCount().getCount().getValue();

				//get the type
				DataComponentPropertyType elemType = array.getElementType();

				/*
				 * Build event structure.
				 */
				if (elemType.isSetAbstractDataRecord()) {
					AbstractDataRecordType adr = elemType.getAbstractDataRecord();

					//SimpleDataRecord
					if (adr instanceof SimpleDataRecordType) {
						SimpleDataRecordType sdr = (SimpleDataRecordType) adr;
						for (int i = 0; i < sdr.getFieldArray().length; i++) {
							PhenomenonUomModel pum = getPhenomenonAndUomFromScalar(sdr.getFieldArray(i));
							pum.setPosition(i);
							pums.add(pum);
						}
					}

					else if (adr instanceof DataRecordType) {
						DataRecordType sdr = (DataRecordType) adr;
						for (int i = 0; i < sdr.getFieldArray().length; i++) {
							PhenomenonUomModel pum = getPhenomenonAndUomFromComponent(sdr.getFieldArray(i));
							pum.setPosition(i);
							pums.add(pum);
						}
					}
				}


				/*
				 * Fill MapEvent with data.
				 */
				final TextBlock textblock = array.getEncoding().getTextBlock();
				final String blockS = textblock.getBlockSeparator();
				final String tokenS = textblock.getTokenSeparator();
				final String decimalS = textblock.getDecimalSeparator();
				
				Element elem = (Element) array.getValues().getDomNode();
				String values = XmlUtils.toString(elem.getFirstChild());
				
				StringTokenizer line = new StringTokenizer(values, blockS);
				
				String current = null;
				int count = 0;
				int i = 0;
				while (line.hasMoreTokens()) {
					i = 0;
					MapEvent newEvent = new MapEvent(0, 0);
					current = line.nextToken();
					StringTokenizer fields = new StringTokenizer(current, tokenS);
					
					while (fields.hasMoreTokens()) {
						
						PhenomenonUomModel pum = pums.get(i);
						if (pum.isTime()) {
							hasTimestamp = true;
						}
						
						if (i != pum.getPosition()) {
							getLog().warning("Probably using wrong phenomenon for values! Recheck " +
									"code.");
						}
						current = fields.nextToken();
						
						if (pum.isUcum()) {
							//got value for conversion
							current.replaceAll(decimalS, ".");
							Double value = Double.parseDouble(current);
							Object[] conv = converter.convert(pum.getUom(), pum.getPhenomenon(), value);
							
							newEvent.put(pum.getPhenomenon(), conv[1]);
						} 
						else if (pum.isTime()) {
							long timeStamp = getTimestamp(current, pum);
							newEvent.put(MapEvent.START_KEY, timeStamp);
							newEvent.put(MapEvent.END_KEY, timeStamp);
						}
						else {
							newEvent.put(pum.getPhenomenon(), current);
						}
						
						
						i++;
					}
					
					events.add(newEvent);
					count++;
				}
				
				if (count != elemCount.intValue()) {
					getLog().info("Unexpected element count: expected "+ elemCount +", got "+
							count +".");
				}

			}

		} else {
			throw new Exception("Expected MessageStructure: XmlAnyType");
		}
		
		/*
		 * parse time from samplingTime
		 */
		if (!hasTimestamp) {
			
			TimeObjectPropertyType topt = obs.getSamplingTime();
			AbstractTimeObjectType to = topt.getTimeObject();
			
			if (to instanceof TimeInstantType) {
				TimeInstantType tit = (TimeInstantType) to;
				TimePositionType tp = tit.getTimePosition();
				long timeStamp = getTimestamp(tp.getStringValue(), null);
				
				//put time
				for (MapEvent me : events) {
					me.put(MapEvent.START_KEY, timeStamp);
					me.put(MapEvent.END_KEY, timeStamp);
				}
			}
			
			else if (to instanceof TimePeriodType) {
				TimePeriodType tpt = (TimePeriodType) to;
				TimePositionType bp = tpt.getBeginPosition();
				TimePositionType ep = tpt.getEndPosition();
				
				long begin = getTimestamp(bp.getStringValue(), null);
				long end = getTimestamp(ep.getStringValue(), null);
				
				//put time
				for (MapEvent me : events) {
					me.put(MapEvent.START_KEY, begin);
					me.put(MapEvent.END_KEY, end);
				}
			}
		}
		
		/*
		 * parse the sensor-id from procedure
		 */
		ProcessPropertyType proc = obs.getProcedure();
		String sensorId = "";
		if (proc.isSetHref()) {
			sensorId = proc.getHref();
		}
		
		//put sensor-id
		for (MapEvent me : events) {
			me.put(MapEvent.SENSORID_KEY, sensorId);
		}
		
		/*
		 * TODO parse observerd property
		 */
		
		
		/*
		 * parse featureOfInterest -> Geometry
		 */
		FeaturePropertyType foi = obs.getFeatureOfInterest();
		AbstractFeatureType feature = foi.getFeature();
		
		Geometry geom = null;
		if (feature instanceof FeatureCollectionType) {
			FeatureCollectionType fct = (FeatureCollectionType) feature;
			FeaturePropertyType[] fmarray = fct.getFeatureMemberArray();

			for (FeaturePropertyType fpt : fmarray) {
				
				AbstractFeatureType member = fpt.getFeature();
	            
	            if (member instanceof SamplingPointPropertyType) {
	               //TODO parse...
	            }
	            
	            else if (member instanceof SamplingPointType) {
	            	SamplingPointType spt = (SamplingPointType) member;
	            	geom = parsePositionFromPoint(spt);
	            }
	            
				else if (member instanceof SamplingFeatureType) {
					//TODO parse SamplingFeature
				}
	            
			}
		}
		
		//put geometry
		if (geom != null) {
			for (MapEvent me : events) {
				me.put(MapEvent.GEOMETRY_KEY, geom);
			}
		}

		return events;
	}


	private Geometry parsePositionFromPoint(SamplingPointType spt) throws com.vividsolutions.jts.io.ParseException {
		if (spt.getPosition().isSetPoint()) {
			PointType point = spt.getPosition().getPoint();
			if (point.isSetPos()) {
				WKTReader reader = new WKTReader();
				return reader.read("POINT("+ point.getPos().getStringValue() + ")");
			}
		}
		
		else if (spt.isSetBoundedBy()) {
			BoundingShapeType bb = spt.getBoundedBy();
			EnvelopeType env = bb.getEnvelope();
			return getGeometryFromEnvelope(env);
		}
		
		return null;
	}

	private Geometry getGeometryFromEnvelope(EnvelopeType et) throws com.vividsolutions.jts.io.ParseException {
		Element elem = (Element) et.getUpperCorner().getDomNode();
		String ucString = XmlUtils.toString(elem.getFirstChild());
		String[] uc = ucString.split(" ");
		
		int i = 0;
		for (String string : uc) {
			string = string.trim();
			uc[i] = string;
			i++;
		}

		//get lower corner and split at " "
		elem = (Element) et.getLowerCorner().getDomNode();
		String lcString = XmlUtils.toString(elem.getFirstChild());
		String[] lc = lcString.split(" ");

		i = 0;
		for (String string : lc) {
			string = string.trim();
			lc[i] = string;
			i++;
		}
		
		String wktString = "POLYGON(("+ lc[0] +" "+ lc[1] +", "+ lc[0] +" "+ uc[1] +", "+ uc[0] +" "+ uc[1] +
				", "+ uc[0] +" "+ lc[1] +", "+ lc[0] +" "+ lc[1]+ "))";
		
		WKTReader wktReader = new WKTReader();
		return wktReader.read(wktString);		
	}

	private PhenomenonUomModel getPhenomenonAndUomFromScalar(AnyScalarPropertyType scalar) throws Exception {

		PhenomenonUomModel result = new PhenomenonUomModel();
		final String phen;
		final String uom;

		if (scalar.isSetBoolean()) {
			phen = scalar.getBoolean().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = "";
		} else if (scalar.isSetQuantity()) {
			phen = scalar.getQuantity().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = getUom(scalar.getQuantity().getUom(), result);
		} else if (scalar.isSetCount()) {
			phen = scalar.getCount().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = "";
		} else if (scalar.isSetTime()) {
			phen = scalar.getTime().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = getUom(scalar.getTime().getUom(), result);
			result.setTime(true);
		} else if (scalar.isSetText()) {
			uom = "";
			phen = scalar.getText().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
		} else if(scalar.isSetCategory()) {
			uom = "";
			phen = scalar.getCategory().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
		} else {
			return null;
		}

		result.setPhenomenon(phen);
		result.setUom(uom);
		
		return result;
	}
	
	
	private PhenomenonUomModel getPhenomenonAndUomFromComponent(
			DataComponentPropertyType component) throws Exception {
		
		PhenomenonUomModel result = new PhenomenonUomModel();
		final String phen;
		final String uom;
		
		if (component.isSetBoolean()) {
			phen = component.getBoolean().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = "";
		} else if (component.isSetQuantity()) {
			phen = component.getQuantity().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = getUom(component.getQuantity().getUom(), result);
		} else if (component.isSetCount()) {
			phen = component.getCount().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = "";
		} else if (component.isSetTime()) {
			phen = component.getTime().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
			uom = getUom(component.getTime().getUom(), result);
			result.setTime(true);
		} else if (component.isSetText()) {
			uom = "";
			phen = component.getText().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
		} else if(component.isSetCategory()) {
			uom = "";
			phen = component.getCategory().getDefinition().replaceAll(":", "__").replaceAll("\\.", "_");
		} else {
			throw new Exception("the DataComponentPropertyType is an unsupported type:" + component.getType());
		}
		
		result.setPhenomenon(phen);
		result.setUom(uom);
		
		return result;
	}


	private String getUom(UomPropertyType uom, PhenomenonUomModel result) {
		if (uom != null) {
			if (uom.isSetHref()) {
				return uom.getHref();
			}
			
			else if (uom.isSetCode()) {
				result.setUcum(true);
				return uom.getCode();
			}
			
			//TODO implement UnitDefinition
		}
		
		return "";
	}



	private long getTimestamp(String current, PhenomenonUomModel pum) throws ParseException {
		if (pum != null && pum.getUom().equals("")) {
			//treat as normal timestamp according ISO8601
			DateTime dt = new DateTime(current);
			return dt.getMillis();
		}
		else {
			//try also joda time
			DateTime dt = new DateTime(current);
			return dt.getMillis();
		}
	}
	
	
	
	private Logger getLog() {
		return logger;
	}
	
	

	/**
	 * Helper class for simple access of phenomenon with uom.
	 */
	private final class PhenomenonUomModel {

		private String phenomenon;
		private String uom;
		private boolean isUcum = false;
		private boolean isTime = false;
		private int position = -1;

		public PhenomenonUomModel(String phenomenon, String uom) {
			super();
			this.phenomenon = phenomenon;
			this.uom = uom;
		}

		public PhenomenonUomModel() {
		}

		public int getPosition() {
			return position;
		}

		public void setPosition(int position) {
			this.position = position;
		}

		public String getPhenomenon() {
			return phenomenon;
		}

		public String getUom() {
			return uom;
		}

		public void setPhenomenon(String phenomenon) {
			this.phenomenon = phenomenon;
		}

		public void setUom(String uom) {
			this.uom = uom;
		}

		public boolean isUcum() {
			return isUcum;
		}

		public void setUcum(boolean isUcum) {
			this.isUcum = isUcum;
		}

		public boolean isTime() {
			return isTime;
		}

		public void setTime(boolean isTime) {
			this.isTime = isTime;
		}

	}
	

}


