/***************************************************************
Copyright 2007 52North Initiative for Geospatial Open Source Software GmbH

 Author: jtheuer, University of Muenster

 Contact: Andreas Wytzisk, 
 52North Initiative for Geospatial Open Source SoftwareGmbH, 
 Martin-Luther-King-Weg 24,
 48155 Muenster, Germany, 
 info@52north.org

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2 as published by the Free Software Foundation.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; even without the implied WARRANTY OF
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program (see gnu-gpl v2.txt). If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA or visit the Free
 Software Foundations web page, http://www.fsf.org.

 ***************************************************************/

package org.n52.swe.sas.muse.wsbr.soaphandler;

import java.text.ParseException;
import java.util.*;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.notification.faults.InvalidTopicExpressionFault;
import org.apache.muse.ws.resource.basefaults.BaseFault;
import org.apache.muse.ws.resource.ext.faults.InvalidMessageFormatFault;
import org.n52.swe.sas.muse.wsbr.*;
import org.n52.swe.sas.muse.wsrf.WsrfRlConstants;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * @author Jan Torben Heuer <jan.heuer@uni-muenster.de>
 * 
 */
public class Publisher implements XmlSerializable {
	/* automatically generated Logger */@SuppressWarnings("unused")
	private static final Logger LOGGER = Logger.getLogger(Publisher.class.getName());

	private EndpointReference publicationEndpoint;
	private Date terminationTime;
	private List<QName> topics;
	private boolean demand;
	private Element sensorML;

	/**
	 * Creates a new Publisher which holds the reference to the new
	 * {@link PublisherEndpoint} instance.
	 * 
	 * @param xml
	 * @throws BaseFault
	 */
	public Publisher(Element xml) throws BaseFault {
		Element eprXML = XmlUtils.getElement(xml, WsbrConstants.TOPIC);

		if (eprXML == null)
			throw new NullPointerException("NullTopicElement");

		/* sensorml */
		sensorML = XmlUtils.getElement(xml,SensorMLConstants.SENSORML);
		
		/* lifetime */
		Element timeXML = XmlUtils.getElement(xml, WsrfRlConstants.REQUESTED_LIFETIME_DURATION);
		
		/* topics */
		setTopic(xml);

		/* demand */
		Element demandElement = XmlUtils.getElement(xml, WsbrConstants.DEMAND);
		//TODO: fix? what are possible values?
		demand = demandElement == null ? false : true;
		
		try {
			if (timeXML != null)
				terminationTime = XmlUtils.getDate(timeXML);
		} catch (ParseException error) {
			throw new InvalidMessageFormatFault("InvalidTerminationTime",error);
		}
	}

	/**
	 * For direct instantiation of new Publisher serializer instances.
	 * @param publisherReference
	 * @param topic
	 * @param onDemand
	 * @param initialTerminationTime
	 * @throws BaseFault 
	 */
	public Publisher(EndpointReference publisherReference, Element[] topic, boolean onDemand, Date initialTerminationTime) throws BaseFault {
		this.publicationEndpoint = publisherReference;
		setTopic(topic);
		this.demand = onDemand;
		this.terminationTime = initialTerminationTime;
	}

	/**
	 * @return an Object array of arguments suitable for invoking {@link RegisterPublisher#registerPublisher(EndpointReference, QName[], boolean, Date)}
	 */
	public Object[] toArray() {
		/* convert OName list to QName array */
		QName[] topicarray = new QName[topics.size()];
		topicarray = topics.toArray(topicarray);
		return new Object[] { null, topicarray, demand, getTerminationTime(), sensorML };
	}
	
	/**
	 * sets the topics
	 * @param rootelement the root topic xml element
	 * @throws InvalidTopicExpressionFault 
	 */
	private void setTopic(Element rootelement) throws InvalidTopicExpressionFault {
		Element[] alltopics = XmlUtils.getAllElements(rootelement,WsbrConstants.NAMESPACE_URI, "Topic"); 
		setTopic(alltopics);
		
	}

	/**
	 * sets the topic
	 * @param newTopics an array of topic xml elements
	 * @throws InvalidTopicExpressionFault 
	 */
	private void setTopic(Element[] alltopics) throws InvalidTopicExpressionFault {
		topics = new ArrayList<QName>(alltopics.length);
		for(Element t: alltopics) {
			if(t.getAttribute("dialect").equals("http://docs.oasis-open.org/wsn/t-1/TopicExpression/Simple")) {
				topics.add(XmlUtils.getQName(t));
			} else {
				throw new InvalidTopicExpressionFault(t.getAttribute("dialect") + " is not a supported dialect. Only http://docs.oasis-open.org/wsn/t-1/TopicExpression/Simple is allowed");
			}
		}
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.apache.muse.util.xml.XmlSerializable#toXML()
	 */
	public Element toXML() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.apache.muse.util.xml.XmlSerializable#toXML(org.w3c.dom.Document)
	 */
	public Element toXML(Document factory) {
		return null;
	}

	public Date getTerminationTime() {
		return terminationTime;
	}

	public EndpointReference getPublicationEndpoint() {
		return publicationEndpoint;
	}

}
