package org.n52.swe.sas.muse;

import org.apache.muse.core.Environment;
import org.apache.muse.core.platform.AbstractIsolationLayer;
import org.apache.muse.core.routing.ResourceRouter;
import org.apache.muse.util.LoggingUtils;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.MessageHeaders;
import org.apache.muse.ws.addressing.WsaConstants;
import org.apache.muse.ws.addressing.soap.*;
import org.w3c.dom.*;

public class InlineIsolationLayer extends AbstractIsolationLayer {

	private Environment environment;
	
	@Override
	protected Environment createEnvironment() {
		environment = new InlineEnvironment(this);
		return environment;
	}

	/**
	 * Handles muse requests
	 * 
	 * @param request
	 * @return muse response
	 */
	public Document handleRequest(Document request) {
		Element soap = XmlUtils.getFirstElement(request);
		Element header = XmlUtils.getElement(soap, SoapConstants.HEADER_QNAME);
		Element body = XmlUtils.getElement(soap, SoapConstants.BODY_QNAME);

		if (header == null)
			throw new RuntimeException("Invalid SOAP envelope: no header element.");

		if (body == null)
			throw new RuntimeException("Invalid SOAP envelope: no body element.");

		Element requestData = XmlUtils.getFirstElement(body);

		ResourceRouter router = getRouter();
		Environment env = router.getEnvironment();

		MessageHeaders addressing = null;

		try {
			//
			// WS-A info provides resource context for this request
			//
			addressing = new MessageHeaders(header);
			env.addAddressingContext(addressing);
		}

		catch (Throwable error) {
			LoggingUtils.logError(router.getLog(), error);
		}

		if (!hasFailedToInitialize())
			LoggingUtils.logMessage(router.getLog(), request, true);

		//
		// actually perform the operation with the given parameters
		//
		Element result = router.invoke(requestData);

		//
		// the standard response wsa:Action name is Action + "Response"
		//
		MessageHeaders replyAddressing = null;

		/*
		 * Difference from mini platform - CTw. We need to check what the error
		 * was. If is due to invalid wsa then we need to know
		 * 
		 */
		if (result != null && SoapUtils.isFault(result)) {
			SoapFault fault = new SoapFault(result);

			if (WsaConstants.DESTINATION_UNREACHABLE_FAULT_QNAME.equals(fault.getSubCode())) {
				// we couldn't find a matching resource, therefore give up.
				System.err.println(fault);
				env.removeAddressingContext();
				return null;
			}
		}

		//
		// send back fault headers on exception
		//
		if (result != null && SoapUtils.isFault(result)) {
			replyAddressing = addressing.createFaultHeaders();
			System.err.println("had fault " + XmlUtils.toString(result));
		} else
			replyAddressing = addressing.createReplyHeaders();

		//
		// DONE - must be sure to remove the request context, or
		// we'll have a memory leak
		//
		env.removeAddressingContext();

		//
		// import all of the headers into the response envelope...
		//        
		Element replyXML = replyAddressing.toXML();
		Element[] children = XmlUtils.getAllElements(replyXML);

		Document response = XmlUtils.createDocument();

		soap = XmlUtils.createElement(response, SoapConstants.ENVELOPE_QNAME);
		response.appendChild(soap);

		header = XmlUtils.createElement(response, SoapConstants.HEADER_QNAME);
		soap.appendChild(header);

		for (int n = 0; n < children.length; ++n) {
			Node next = response.importNode(children[n], true);
			header.appendChild(next);
		}

		//
		// add the result (valid or fault) to the SOAP body...
		//
		body = XmlUtils.createElement(response, SoapConstants.BODY_QNAME);
		soap.appendChild(body);

		if (result != null) {
			result = (Element) response.importNode(result, true);
			body.appendChild(result);
		}

		if (!hasFailedToInitialize())
			LoggingUtils.logMessage(router.getLog(), response, false);

		return response;
	}

	public Environment getEnvironment() {
		if(environment == null) {
			createEnvironment();
		}
		return environment;
	}

}
